"""
A basic python script that demonstrates how to use the FST8 reader, writer, and wrapper in a purely
python setting. These functions are constructed to provide a simple interface for controlling FAST
programmatically with minimal additional dependencies.
"""
# Hacky way of doing relative imports
from __future__ import print_function
import os, sys, time
import multiprocessing as mp
# sys.path.insert(0, os.path.abspath(".."))

from wisdem.aeroelasticse.FAST_reader import InputReader_Common, InputReader_OpenFAST, InputReader_FAST7
from wisdem.aeroelasticse.FAST_writer import InputWriter_Common, InputWriter_OpenFAST, InputWriter_FAST7
from wisdem.aeroelasticse.FAST_wrapper import FastWrapper
from wisdem.aeroelasticse.FAST_post import return_timeseries

import numpy as np

class runFAST_pywrapper(object):

    def __init__(self, **kwargs):
        self.FAST_ver = 'OPENFAST' #(FAST7, FAST8, OPENFAST)

        self.FAST_exe = None
        self.FAST_InputFile = None
        self.FAST_directory = None
        self.FAST_runDirectory = None
        self.FAST_namingOut = None
        self.read_yaml = False
        self.write_yaml = False
        self.fst_vt = {}
        self.case = {}                  # dictionary of variable values to change
        self.channels = {}              # dictionary of output channels to change
        self.debug_level   = 0
        self.dev_branch = False

        # Optional population class attributes from key word arguments
        for (k, w) in kwargs.items():
            try:
                setattr(self, k, w)
            except:
                pass

        super(runFAST_pywrapper, self).__init__()

    def execute(self):
        # FAST version specific initialization
        if self.FAST_ver.lower() == 'fast7':
            reader = InputReader_FAST7(FAST_ver=self.FAST_ver)
            writer = InputWriter_FAST7(FAST_ver=self.FAST_ver)
        elif self.FAST_ver.lower() in ['fast8','openfast']:
            reader = InputReader_OpenFAST(FAST_ver=self.FAST_ver)
            writer = InputWriter_OpenFAST(FAST_ver=self.FAST_ver)
        wrapper = FastWrapper(FAST_ver=self.FAST_ver, debug_level=self.debug_level)

        # Read input model, FAST files or Yaml
        if self.fst_vt == {}:
            if self.read_yaml:
                reader.FAST_yamlfile = self.FAST_yamlfile_in
                reader.read_yaml()
            else:
                reader.FAST_InputFile = self.FAST_InputFile
                reader.FAST_directory = self.FAST_directory
                reader.dev_branch = self.dev_branch
                reader.execute()
        
            # Initialize writer variables with input model
            writer.fst_vt = reader.fst_vt
        else:
            writer.fst_vt = self.fst_vt
        writer.FAST_runDirectory = self.FAST_runDirectory
        writer.FAST_namingOut = self.FAST_namingOut
        writer.dev_branch = self.dev_branch
        # Make any case specific variable changes
        if self.case:
            writer.update(fst_update=self.case)
        # Modify any specified output channels
        if self.channels:
            writer.update_outlist(self.channels)
        # Write out FAST model
        writer.execute()
        if self.write_yaml:
            writer.FAST_yamlfile = self.FAST_yamlfile_out
            writer.write_yaml()

        # Run FAST
        wrapper.FAST_exe = self.FAST_exe
        wrapper.FAST_InputFile = os.path.split(writer.FAST_InputFileOut)[1]
        wrapper.FAST_directory = os.path.split(writer.FAST_InputFileOut)[0]
        wrapper.execute()

        FAST_Output = os.path.join(wrapper.FAST_directory, wrapper.FAST_InputFile[:-3]+'outb')
        return FAST_Output

class runFAST_pywrapper_batch(object):

    def __init__(self, **kwargs):

        self.FAST_ver           = 'OpenFAST'
        self.FAST_exe           = None
        self.FAST_InputFile     = None
        self.FAST_directory     = None
        self.FAST_runDirectory  = None
        self.debug_level        = 0
        self.dev_branch         = False

        self.read_yaml          = False
        self.FAST_yamlfile_in   = ''
        self.fst_vt             = {}
        self.write_yaml         = False
        self.FAST_yamlfile_out  = ''

        self.case_list          = []
        self.case_name_list     = []
        self.channels           = {}

        self.post               = None

        # Optional population of class attributes from key word arguments
        for (k, w) in kwargs.items():
            try:
                setattr(self, k, w)
            except:
                pass

        super(runFAST_pywrapper_batch, self).__init__()

        
    def run_serial(self):
        # Run batch serially

        if not os.path.exists(self.FAST_runDirectory):
            os.makedirs(self.FAST_runDirectory)

        out = [None]*len(self.case_list)
        for i, (case, case_name) in enumerate(zip(self.case_list, self.case_name_list)):
            out[i] = eval(case, case_name, self.FAST_ver, self.FAST_exe, self.FAST_runDirectory, self.FAST_InputFile, self.FAST_directory, self.read_yaml, self.FAST_yamlfile_in, self.fst_vt, self.write_yaml, self.FAST_yamlfile_out, self.channels, self.debug_level, self.dev_branch, self.post)

        return out

    def run_multi(self, cores=None):
        # Run cases in parallel, threaded with multiprocessing module

        if not os.path.exists(self.FAST_runDirectory):
            os.makedirs(self.FAST_runDirectory)

        if not cores:
            cores = mp.cpu_count()
        pool = mp.Pool(cores)

        case_data_all = []
        for i in range(len(self.case_list)):
            case_data = []
            case_data.append(self.case_list[i])
            case_data.append(self.case_name_list[i])
            case_data.append(self.FAST_ver)
            case_data.append(self.FAST_exe)
            case_data.append(self.FAST_runDirectory)
            case_data.append(self.FAST_InputFile)
            case_data.append(self.FAST_directory)
            case_data.append(self.read_yaml)
            case_data.append(self.FAST_yamlfile_in)
            case_data.append(self.fst_vt)
            case_data.append(self.write_yaml)
            case_data.append(self.FAST_yamlfile_out)
            case_data.append(self.channels)
            case_data.append(self.debug_level)
            case_data.append(self.dev_branch)
            case_data.append(self.post)

            case_data_all.append(case_data)

        output = pool.map(eval_multi, case_data_all)
        pool.close()
        pool.join()

        return output

    def run_mpi(self, mpi_comm_map_down):
        # Run in parallel with mpi
        from mpi4py import MPI

        # mpi comm management
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        sub_ranks = mpi_comm_map_down[rank]
        size = len(sub_ranks)

        N_cases = len(self.case_list)
        N_loops = int(np.ceil(float(N_cases)/float(size)))
        
        # file management
        if not os.path.exists(self.FAST_runDirectory) and rank == 0:
            os.makedirs(self.FAST_runDirectory)

        case_data_all = []
        for i in range(N_cases):
            case_data = []
            case_data.append(self.case_list[i])
            case_data.append(self.case_name_list[i])
            case_data.append(self.FAST_ver)
            case_data.append(self.FAST_exe)
            case_data.append(self.FAST_runDirectory)
            case_data.append(self.FAST_InputFile)
            case_data.append(self.FAST_directory)
            case_data.append(self.read_yaml)
            case_data.append(self.FAST_yamlfile_in)
            case_data.append(self.fst_vt)
            case_data.append(self.write_yaml)
            case_data.append(self.FAST_yamlfile_out)
            case_data.append(self.channels)
            case_data.append(self.debug_level)
            case_data.append(self.dev_branch)
            case_data.append(self.post)

            case_data_all.append(case_data)

        output = []
        for i in range(N_loops):
            idx_s    = i*size
            idx_e    = min((i+1)*size, N_cases)

            for j, case_data in enumerate(case_data_all[idx_s:idx_e]):
                data   = [eval_multi, case_data]
                rank_j = sub_ranks[j]
                comm.send(data, dest=rank_j, tag=0)

            # for rank_j in sub_ranks:
            for j, case_data in enumerate(case_data_all[idx_s:idx_e]):
                rank_j = sub_ranks[j]
                data_out = comm.recv(source=rank_j, tag=1)
                output.append(data_out)

        return output


    # def run_mpi(self, comm=None):
    #     # Run in parallel with mpi
    #     from mpi4py import MPI

    #     # mpi comm management
    #     if not comm:
    #         comm = MPI.COMM_WORLD
    #     size = comm.Get_size()
    #     rank = comm.Get_rank()

    #     N_cases = len(self.case_list)
    #     N_loops = int(np.ceil(float(N_cases)/float(size)))
        
    #     # file management
    #     if not os.path.exists(self.FAST_runDirectory) and rank == 0:
    #         os.makedirs(self.FAST_runDirectory)

    #     if rank == 0:
    #         case_data_all = []
    #         for i in range(N_cases):
    #             case_data = []
    #             case_data.append(self.case_list[i])
    #             case_data.append(self.case_name_list[i])
    #             case_data.append(self.FAST_ver)
    #             case_data.append(self.FAST_exe)
    #             case_data.append(self.FAST_runDirectory)
    #             case_data.append(self.FAST_InputFile)
    #             case_data.append(self.FAST_directory)
    #             case_data.append(self.read_yaml)
    #             case_data.append(self.FAST_yamlfile_in)
    #             case_data.append(self.fst_vt)
    #             case_data.append(self.write_yaml)
    #             case_data.append(self.FAST_yamlfile_out)
    #             case_data.append(self.channels)
    #             case_data.append(self.debug_level)
    #             case_data.append(self.dev_branch)
    #             case_data.append(self.post)

    #             case_data_all.append(case_data)
    #     else:
    #         case_data_all = []

    #     output = []
    #     for i in range(N_loops):
    #         # if # of cases left to run is less than comm size, split comm
    #         n_resid = N_cases - i*size
    #         if n_resid < size: 
    #             split_comm = True
    #             color = np.zeros(size)
    #             for i in range(n_resid):
    #                 color[i] = 1
    #             color = [int(j) for j in color]
    #             comm_i  = MPI.COMM_WORLD.Split(color, 1)
    #         else:
    #             split_comm = False
    #             comm_i = comm

    #         # position in case list
    #         idx_s  = i*size
    #         idx_e  = min((i+1)*size, N_cases)

    #         # scatter out cases
    #         if split_comm:
    #             if color[rank] == 1:
    #                 case_data_i = comm_i.scatter(case_data_all[idx_s:idx_e], root=0)    
    #         else:
    #             case_data_i = comm_i.scatter(case_data_all[idx_s:idx_e], root=0)
            
    #         # eval
    #         out = eval_multi(case_data_i)

    #         # gather results
    #         if split_comm:
    #             if color[rank] == 1:
    #                 output_i = comm_i.gather(out, root=0)
    #         else:
    #             output_i = comm_i.gather(out, root=0)

    #         if rank == 0:
    #             output.extend(output_i)

        # return output



def eval(case, case_name, FAST_ver, FAST_exe, FAST_runDirectory, FAST_InputFile, FAST_directory, read_yaml, FAST_yamlfile_in, fst_vt, write_yaml, FAST_yamlfile_out, channels, debug_level, dev_branch, post):
    # Batch FAST pyWrapper call, as a function outside the runFAST_pywrapper_batch class for pickle-ablility

    fast = runFAST_pywrapper(FAST_ver=FAST_ver)
    fast.FAST_exe           = FAST_exe
    fast.FAST_InputFile     = FAST_InputFile
    fast.FAST_directory     = FAST_directory
    fast.FAST_runDirectory  = FAST_runDirectory
    fast.dev_branch         = dev_branch

    fast.read_yaml          = read_yaml
    fast.FAST_yamlfile_in   = FAST_yamlfile_in
    fast.fst_vt             = fst_vt
    fast.write_yaml         = write_yaml
    fast.FAST_yamlfile_out  = FAST_yamlfile_out

    fast.FAST_namingOut     = case_name
    fast.case               = case
    fast.channels           = channels
    fast.debug_level        = debug_level

    FAST_Output = fast.execute()

    # Post process
    if post:
        out = post(FAST_Output)
    else:
        out = []

    return out

def eval_multi(data):
    # helper function for running with multiprocessing.Pool.map
    # converts list of arguement values to arguments
    return eval(data[0], data[1], data[2], data[3], data[4], data[5], data[6], data[7], data[8], data[9], data[10], data[11], data[12], data[13], data[14], data[15])

def example_runFAST_pywrapper_batch():
    """ 
    Example of running a batch of cases, in serial or in parallel
    """
    fastBatch = runFAST_pywrapper_batch(FAST_ver='OpenFAST', dev_branch=True)

    # fastBatch.FAST_exe = 'C:/Users/egaertne/WT_Codes/openfast/build/glue-codes/fast/openfast.exe'   # Path to executable
   # fastBatch.FAST_InputFile = '5MW_Land_DLL_WTurb.fst'   # FAST input file (ext=.fst)
    # fastBatch.FAST_directory = 'C:/Users/egaertne/WT_Codes/models/openfast/glue-codes/fast/5MW_Land_DLL_WTurb'   # Path to fst directory files
    # fastBatch.FAST_runDirectory = 'temp/OpenFAST'
    # fastBatch.debug_level = 2
    fastBatch.FAST_exe          = '/projects/windse/importance_sampling/WT_Codes/openfast/build/glue-codes/openfast/openfast'   # Path to executable
    fastBatch.FAST_InputFile    = '5MW_Land_DLL_WTurb.fst'   # FAST input file (ext=.fst)
    fastBatch.FAST_directory    = "/projects/windse/importance_sampling/WISDEM/xloads_tc/templates/openfast/5MW_Land_DLL_WTurb-Shutdown"   # Path to fst directory files
    fastBatch.FAST_runDirectory = 'temp/OpenFAST'
    fastBatch.debug_level       = 2
    fastBatch.post              = return_timeseries


    ## Define case list explicitly
    # case_list = [{}, {}]
    # case_list[0]['Fst', 'TMax'] = 4.
    # case_list[1]['Fst', 'TMax'] = 5.
    # case_name_list = ['test01', 'test02']

    ## Generate case list using General Case Generator
    ## Specify several variables that change independently or collectly
    case_inputs = {}
    case_inputs[("Fst","TMax")] = {'vals':[5.], 'group':0}
    case_inputs[("InflowWind","WindType")] = {'vals':[1], 'group':0}
    case_inputs[("Fst","OutFileFmt")] = {'vals':[2], 'group':0}
    case_inputs[("InflowWind","HWindSpeed")] = {'vals':[8., 9., 10., 11., 12.], 'group':1}
    case_inputs[("ElastoDyn","RotSpeed")] = {'vals':[9.156, 10.296, 11.431, 11.89, 12.1], 'group':1}
    case_inputs[("ElastoDyn","BlPitch1")] = {'vals':[0., 0., 0., 0., 3.823], 'group':1}
    case_inputs[("ElastoDyn","BlPitch2")] = case_inputs[("ElastoDyn","BlPitch1")]
    case_inputs[("ElastoDyn","BlPitch3")] = case_inputs[("ElastoDyn","BlPitch1")]
    case_inputs[("ElastoDyn","GenDOF")] = {'vals':['True','False'], 'group':2}
    
    from CaseGen_General import CaseGen_General
    case_list, case_name_list = CaseGen_General(case_inputs, dir_matrix=fastBatch.FAST_runDirectory, namebase='testing')

    fastBatch.case_list = case_list
    fastBatch.case_name_list = case_name_list

    # fastBatch.run_serial()
    # fastBatch.run_multi(2)
    fastBatch.run_mpi()


def example_runFAST_CaseGenIEC():

    from CaseGen_IEC import CaseGen_IEC
    iec = CaseGen_IEC()

    # Turbine Data
    iec.init_cond = {} # can leave as {} if data not available
    iec.init_cond[("ElastoDyn","RotSpeed")] = {'U':[3., 4., 5., 6., 7., 8., 9., 10., 11., 12., 13., 14., 15., 16., 17., 18., 19., 20., 21., 22., 23., 24., 25]}
    iec.init_cond[("ElastoDyn","RotSpeed")]['val'] = [6.972, 7.183, 7.506, 7.942, 8.469, 9.156, 10.296, 11.431, 11.89, 12.1, 12.1, 12.1, 12.1, 12.1, 12.1, 12.1, 12.1, 12.1, 12.1, 12.1, 12.1, 12.1, 12.1]
    iec.init_cond[("ElastoDyn","BlPitch1")] = {'U':[3., 4., 5., 6., 7., 8., 9., 10., 11., 12., 13., 14., 15., 16., 17., 18., 19., 20., 21., 22., 23., 24., 25]}
    iec.init_cond[("ElastoDyn","BlPitch1")]['val'] = [0., 0., 0., 0., 0., 0., 0., 0., 0., 3.823, 6.602, 8.668, 10.450, 12.055, 13.536, 14.920, 16.226, 17.473, 18.699, 19.941, 21.177, 22.347, 23.469]
    iec.init_cond[("ElastoDyn","BlPitch2")] = iec.init_cond[("ElastoDyn","BlPitch1")]
    iec.init_cond[("ElastoDyn","BlPitch3")] = iec.init_cond[("ElastoDyn","BlPitch1")]

    iec.Turbine_Class = 'I' # I, II, III, IV
    iec.Turbulence_Class = 'A'
    iec.D = 126.
    iec.z_hub = 90.

    # DLC inputs
    iec.dlc_inputs = {}
    iec.dlc_inputs['DLC']   = [1.1, 1.5]
    iec.dlc_inputs['U']     = [[8, 9, 10], [12]]
    iec.dlc_inputs['Seeds'] = [[5, 6, 7], []]
    iec.dlc_inputs['Yaw']   = [[], []]

    iec.transient_dir_change        = 'both'  # '+','-','both': sign for transient events in EDC, EWS
    iec.transient_shear_orientation = 'both'  # 'v','h','both': vertical or horizontal shear for EWS

    # Naming, file management, etc
    iec.wind_dir = 'temp/wind'
    iec.case_name_base = 'testing'
    iec.Turbsim_exe = 'C:/Users/egaertne/WT_Codes/Turbsim_v2.00.07/bin/TurbSim_x64.exe'
    iec.debug_level = 2
    iec.parallel_windfile_gen = True
    iec.cores = 4
    iec.run_dir = 'temp/OpenFAST'

    # Run case generator / wind file writing
    case_inputs = {}
    case_inputs[('Fst','OutFileFmt')] = {'vals':[1], 'group':0}
    case_list, case_name_list = iec.execute(case_inputs=case_inputs)

    # Run FAST cases
    fastBatch = runFAST_pywrapper_batch(FAST_ver='OpenFAST')
    fastBatch.FAST_exe = 'C:/Users/egaertne/WT_Codes/openfast/build/glue-codes/fast/openfast.exe'   # Path to executable
    fastBatch.FAST_InputFile = '5MW_Land_DLL_WTurb.fst'   # FAST input file (ext=.fst)
    fastBatch.FAST_directory = 'C:/Users/egaertne/WT_Codes/models/openfast/glue-codes/fast/5MW_Land_DLL_WTurb'   # Path to fst directory files
    fastBatch.FAST_runDirectory = iec.run_dir

    fastBatch.case_list = case_list
    fastBatch.case_name_list = case_name_list
    fastBatch.debug_level = 2

    # fastBatch.run_serial()
    fastBatch.run_multi(4)

    
def example_runFAST_pywrapper():
    """ 
    Example of reading, writing, and running FAST 7, 8 and OpenFAST.
    """

    FAST_ver = 'OpenFAST'
    dev_branch = True
    fast = runFAST_pywrapper(FAST_ver=FAST_ver, dev_branch=dev_branch, debug_level=2)

    if FAST_ver.lower() == 'fast7':
        fast.FAST_exe = 'C:/Users/egaertne/WT_Codes/FAST_v7.02.00d-bjj/FAST.exe'   # Path to executable
        fast.FAST_InputFile = 'Test12.fst'   # FAST input file (ext=.fst)
        fast.FAST_directory = 'C:/Users/egaertne/WT_Codes/models/FAST_v7.02.00d-bjj/CertTest/'   # Path to fst directory files
        fast.FAST_runDirectory = 'temp/FAST7'
        fast.FAST_namingOut = 'test'

    elif FAST_ver.lower() == 'fast8':
        fast.FAST_exe = 'C:/Users/egaertne/WT_Codes/FAST_v8.16.00a-bjj/bin/FAST_Win32.exe'   # Path to executable
        fast.FAST_InputFile = 'NREL5MW_onshore.fst'   # FAST input file (ext=.fst)
        fast.FAST_directory = 'C:/Users/egaertne/WT_Codes/models/FAST_v8.16.00a-bjj/ref/5mw_onshore/'   # Path to fst directory files
        fast.FAST_runDirectory = 'temp/FAST8'
        fast.FAST_namingOut = 'test'

    # elif FAST_ver.lower() == 'openfast':
    #     fast.FAST_exe = 'C:/Users/egaertne/WT_Codes/openfast/build/glue-codes/fast/openfast.exe'   # Path to executable
    #     fast.FAST_InputFile = '5MW_Land_DLL_WTurb.fst'   # FAST input file (ext=.fst)
    #     fast.FAST_directory = 'C:/Users/egaertne/WT_Codes/models/openfast/glue-codes/fast/5MW_Land_DLL_WTurb'   # Path to fst directory files
    #     fast.FAST_runDirectory = 'temp/OpenFAST'
    #     fast.FAST_namingOut = 'test'

    #     fast.read_yaml = False
    #     fast.FAST_yamlfile_in = 'temp/OpenFAST/test.yaml'

    #     fast.write_yaml = False
    #     fast.FAST_yamlfile_out = 'temp/OpenFAST/test.yaml'
    elif FAST_ver.lower() == 'openfast':
        fast.FAST_exe = 'C:/Users/egaertne/WT_Codes/openfast-dev/build/glue-codes/openfast/openfast.exe'   # Path to executable
        # fast.FAST_InputFile = '5MW_Land_DLL_WTurb.fst'   # FAST input file (ext=.fst)
        # fast.FAST_directory = 'C:/Users/egaertne/WT_Codes/models/openfast-dev/r-test/glue-codes/openfast/5MW_Land_DLL_WTurb'   # Path to fst directory files
        fast.FAST_InputFile = '5MW_OC3Spar_DLL_WTurb_WavesIrr.fst'   # FAST input file (ext=.fst)
        fast.FAST_directory = 'C:/Users/egaertne/WT_Codes/models/openfast-dev/r-test/glue-codes/openfast/5MW_OC3Spar_DLL_WTurb_WavesIrr'   # Path to fst directory files
        fast.FAST_runDirectory = 'temp/OpenFAST'
        fast.FAST_namingOut = 'test_run_spar'

        fast.read_yaml = False
        fast.FAST_yamlfile_in = 'temp/OpenFAST/test.yaml'

        fast.write_yaml = False
        fast.FAST_yamlfile_out = 'temp/OpenFAST/test.yaml'

    fast.execute()


if __name__=="__main__":

    # example_runFAST_pywrapper()
    example_runFAST_pywrapper_batch()
    # example_runFAST_CaseGenIEC()


"""EESG.py
Created by Latha Sethuraman, Katherine Dykes.
Copyright (c) NREL. All rights reserved.
Electromagnetic design based on conventional magnetic circuit laws
Structural design based on McDonald's thesis """

from openmdao.api import Group, Problem, ExplicitComponent,ExecComp,IndepVarComp,ScipyOptimizeDriver
import numpy as np
from math import pi, cos, sqrt, radians, sin, exp, log10, log, tan, atan
import sys, os

class EESG(ExplicitComponent):
    """ Estimates overall mass dimensions and Efficiency of Electrically Excited Synchronous generator. """
    
    def setup(self):
        # EESG generator design inputs
        #self.add_input('r_s', val=0.0, units ='m', desc='airgap radius r_s')
        self.add_input('rad_ag', val=0.0, units ='m', desc='airgap radius')
        self.add_input('l_s', val=0.0, units ='m', desc='Stator core length l_s')
        self.add_input('h_s', val=0.0, units ='m', desc='Yoke height h_s')
        self.add_input('tau_p',val=0.0, units ='m', desc='Pole pitch self.tau_p')
        
        self.add_input('machine_rating',val=0.0, units ='W', desc='Machine rating')
        self.add_input('n_nom',val=0.0, units ='rpm', desc='rated speed')
        self.add_input('Torque',val=0.0, units ='N*m', desc='Rated torque ')
        self.add_input('I_f',val=0.0000,units='A',desc='Excitation current')
        self.add_input('N_f',val=0.0,units='A',desc='field turns')
        self.add_input('h_ys',val=0.0, units ='m', desc='Yoke height')
        self.add_input('h_yr',val=0.0, units ='m', desc='rotor yoke height')
        
        # structural design variables
        self.add_input('n_s' ,val=0.0, desc='number of stator arms n_s')
        self.add_input('b_st' , val=0.0, units ='m', desc='arm width b_st')
        self.add_input('d_s',val=0.0,units ='m', desc='arm depth d_s')
        self.add_input('t_ws' ,val=0.0,units ='m', desc='arm depth thickness self.t_wr')
        self.add_input('n_r' ,val=0.0, desc='number of arms n')
        self.add_input('b_r' ,val=0.0,units ='m', desc='arm width b_r')
        self.add_input('d_r' ,val=0.0, units ='m', desc='arm depth d_r')
        self.add_input('t_wr' ,val=0.0, units ='m', desc='arm depth thickness self.t_wr')
        self.add_input('R_o',val=0.0, units ='m',desc='Shaft radius')
        
        # EESG generator design outputs
        
        # Magnetic loading
        self.add_output('B_symax' ,val=0.0, desc='Peak Stator Yoke flux density B_ymax')
        self.add_output('B_tmax',val=0.0, desc='Peak Teeth flux density')
        self.add_output('B_rymax',val=0.0, desc='Peak Rotor yoke flux density')
        self.add_output('B_gfm',val=0.0, desc='Average air gap flux density B_g')
        self.add_output('B_g' ,val=0.0, desc='Peak air gap flux density B_g')
        self.add_output('B_pc',val=0.0, desc='Pole core flux density')
        
        # Stator design
        self.add_output('N_s' ,val=0.0, desc='Number of turns in the stator winding')
        self.add_output('b_s',val=0.0, desc='slot width')
        self.add_output('b_t',val=0.0, desc='tooth width')
        self.add_output('A_Cuscalc',val=0.0, desc='Conductor cross-section mm^2')
        self.add_output('S',val=0.0, desc='Stator slots')
        
        # # Output parameters : Rotor design
        self.add_output('h_p',val=0.0, desc='Pole height')
        self.add_output('b_p',val=0.0, desc='Pole width')
        self.add_output('p',val=0.0, desc='No of pole pairs')
        self.add_output('n_brushes',val=0.0, desc='number of brushes')
        self.add_output('A_Curcalc',val=0.0, desc='Rotor Conductor cross-section')
        
        # Output parameters : Electrical performance
        self.add_output('E_s',val=0.0, desc='Stator phase voltage')
        self.add_output('f',val=0.0, desc='Generator output frequency')
        self.add_output('I_s',val=0.0, desc='Generator output phase current')
        self.add_output('R_s',val=0.0, desc='Stator resistance')
        self.add_output('R_r',val=0.0, desc='Rotor resistance')
        self.add_output('L_m',val=0.0, desc='Stator synchronising inductance')
        self.add_output('J_s',val=0.0, desc='Stator Current density')
        self.add_output('J_f',val=0.0, desc='rotor Current density')
        self.add_output('A_1',val=0.0, desc='Specific current loading')
        self.add_output('Load_mmf_ratio',val=0.0, desc='mmf_ratio')
        
        # Objective functions and output
        self.add_output('Mass',val=0.0, desc='Actual mass')
        self.add_output('K_rad',val=0.0, desc='K_rad')
        self.add_output('Losses',val=0.0, desc='Total loss')
        self.add_output('gen_eff',val=0.0, desc='Generator efficiency')
        
        # Structural performance
        self.add_output('u_Ar',val=0.0, desc='Rotor radial deflection')
        self.add_output('y_Ar',val=0.0, desc='Rotor axial deflection')
        self.add_output('z_A_r',val=0.0, desc='Rotor circumferential deflection')
        self.add_output('u_As',val=0.0, desc='Stator radial deflection')
        self.add_output('y_As',val=0.0, desc='Stator axial deflection')
        self.add_output('z_A_s',val=0.0, desc='Stator circumferential deflection')  
        self.add_output('u_all_r',val=0.0, desc='Allowable radial rotor')
        self.add_output('u_all_s',val=0.0, desc='Allowable radial stator')
        self.add_output('y_all',val=0.0, desc='Allowable axial')
        self.add_output('z_all_s',val=0.0, desc='Allowable circum stator')
        self.add_output('z_all_r',val=0.0, desc='Allowable circum rotor')
        self.add_output('b_all_s',val=0.0, desc='Allowable arm')
        self.add_output('b_all_r',val=0.0, desc='Allowable arm dimensions')
        self.add_output('TC1',val=0.0, desc='Torque constraint')
        self.add_output('TC2',val=0.0, desc='Torque constraint-rotor')
        self.add_output('TC3',val=0.0, desc='Torque constraint-stator')
        
        # Material properties
        self.add_input('rho_Fes',val=0.0,units='kg*m**-3', desc='Structural Steel density ')
        self.add_input('rho_Fe',val=0.0,units='kg*m**-3', desc='Magnetic Steel density ')
        self.add_input('rho_Copper',val=0.0,units='kg*m**-3', desc='Copper density ')
        
        # Mass Outputs
        self.add_output('Copper', val=0.0, units='kg', desc='Copper Mass')
        self.add_output('Iron', val=0.0, units='kg', desc='Electrical Steel Mass')
        self.add_output('Structural_mass', val=0.0, units='kg', desc='Structural Mass')
        
        # Other parameters
        self.add_output('Power_ratio',val=0.0, desc='Power_ratio')
        self.add_output('Slot_aspect_ratio',val=0.0,desc='Stator slot aspect ratio')
        self.add_output('R_out',val=0.0, desc='Outer radius')
        
        # inputs/outputs for interface with drivese
        self.add_input('shaft_cm',val= np.zeros(3), units='m', desc='Main Shaft CM')
        self.add_input('shaft_length',val=0.0, units='m', desc='main shaft length')
        self.add_output('I',val=np.zeros(3),desc='Moments of Inertia for the component [Ixx, Iyy, Izz] around its center of mass')
        self.add_output('cm', val=np.zeros(3),desc='COM [x,y,z]')

        #self.declare_partials('*', '*', method='fd', form='central', step=1e-6)

        
    def compute(self, inputs, outputs):
        # Unpack outputs
        rad_ag               = inputs['rad_ag']
        l_s               = inputs['l_s']
        h_s               = inputs['h_s']
        tau_p             = inputs['tau_p']
        N_f               = inputs['N_f']
        I_f               = inputs['I_f']
        h_ys              = inputs['h_ys']
        h_yr              = inputs['h_yr']
        machine_rating    = inputs['machine_rating']
        n_nom             = inputs['n_nom']
        Torque            = inputs['Torque']
        
        b_st              = inputs['b_st']
        d_s               = inputs['d_s']
        t_ws              = inputs['t_ws']
        n_r               = inputs['n_r']
        n_s               = inputs['n_s']
        b_r               = inputs['b_r']
        d_r               = inputs['d_r']
        t_wr              = inputs['t_wr']
        
        R_o               = inputs['R_o']
        rho_Fe            = inputs['rho_Fe']
        rho_Copper        = inputs['rho_Copper']
        rho_Fes           = inputs['rho_Fes']
        shaft_cm          = inputs['shaft_cm']
        shaft_length      = inputs['shaft_length']
        
        # Assign values to universal constants
        g1     = 9.81                # m / s^2 acceleration due to gravity
        E      = 2e11                # N / m^2 young's modulus
        sigma  = 48.373e3            # shear stress
        mu_0   = pi * 4e-7           # permeability of free space
        phi    = radians(90)
        
        # Assign values to design constants
        h_w       = 0.005
        b_so      = 0.004              # Stator slot opening
        m         = 3                  # number of phases
        q1        = 2                  # no of stator slots per pole per phase
        b_s_tau_s = 0.45               # ratio of slot width to slot pitch
        k_sfil    = 0.65               # Slot fill factor (not used)
        P_Fe0h    = 4                  # specific hysteresis losses W / kg @ 1.5 T @50 Hz
        P_Fe0e    = 1                  # specific eddy losses W / kg @ 1.5 T @50 Hz
        rho_Cu    = 1.8e-8 * 1.4       # resisitivity of copper
        k_fes     = 0.9                # iron fill factor (not used)
        y_tau_p   = 1                  # coil span / pole pitch fullpitch
        k_fillr   = 0.7                # rotor slot fill factor
        k_s       = 0.2                # magnetic saturation factor for iron
        T         = Torque
        cos_phi   = 0.85               # power factor
        
        # back iron thickness for rotor and stator
        t_s = h_ys
        t   = h_yr
        
        # Aspect ratio
        K_rad = l_s / (2 * rad_ag)
        
        ###################################################### Electromagnetic design#############################################
        
        alpha_p = pi / 2 * .7 # (not used)
        dia = 2 * rad_ag             # air gap diameter
        
        # air gap length and minimum values
        g = 0.001 * dia       
        if(g < 0.005):
            g = 0.005
            
        r_r = rad_ag - g                             # rotor radius
        d_se = dia + 2 * h_s + 2 * h_ys           # stator outer diameter (not used)
        p = np.round(pi * dia / (2 * tau_p))      # number of pole pairs
        S = 2 * p*q1 * m                          # number of slots of stator phase winding
        N_conductors = S * 2
        N_s = N_conductors / 2/3                  # Stator turns per phase
        alpha = 180 / S/p                         # electrical angle (not used)
        
        tau_s = pi * dia / S                      # slot pitch
        h_ps = 0.1 * tau_p                        # height of pole shoe
        b_pc = 0.4 * tau_p                        # width of pole core
        h_pc = 0.6 * tau_p                        # height of pole core
        h_p = 0.7 * tau_p                         # pole height
        b_p = h_p
        b_s = tau_s * b_s_tau_s                   # slot width
        Slot_aspect_ratio = h_s / b_s
        b_t = tau_s - b_s                         # tooth width
        
        # Calculating carter factor and effective air gap
        g_a = g
        K_C1 = (tau_s + 10 * g_a) / (tau_s - b_s + 10 * g_a)  # salient pole rotor
        g_1 = K_C1 * g
        
        # calculating angular frequency
        om_m = 2 * pi * n_nom / 60
        om_e = 60
        f    =  n_nom * p / 60
        
        # Slot fill factor according to air gap radius
        
        if (2 * rad_ag>2):
            K_fills = 0.65
        else:
            K_fills = 0.4
            
        # Calculating Stator winding factor    
        
        k_y1 = sin(y_tau_p * pi / 2)                        # chording factor
        k_q1 = sin(pi / 6) / q1 / sin(pi / 6 / q1)          # winding zone factor
        k_wd = k_y1 * k_q1
        
        # Calculating stator winding conductor length, cross-section and resistance
        
        shortpitch = 0
        l_Cus = 2 * N_s * (2 * (tau_p - shortpitch / m/q1) + l_s)   # length of winding
        A_s       = b_s        * (h_s - h_w)
        A_scalc   = b_s * 1000 * (h_s - h_w) * 1000            # cross section in mm^2
        A_Cus     = A_s     * q1 * p * K_fills / N_s
        A_Cuscalc = A_scalc * q1 * p * K_fills / N_s
        R_s = l_Cus * rho_Cu / A_Cus
        
        # field winding design, conductor lenght, cross-section and resistance
        
        N_f       = np.round(N_f)            # rounding the field winding turns to the nearest integer
        I_srated  = machine_rating / (sqrt(3) * 5000 * cos_phi)
        l_pole    = l_s - 0.05 + 0.120  # 50mm smaller than stator and 120mm longer to accommodate end stack
        K_fe      = 0.95                    
        l_pfe     = l_pole * K_fe
        l_Cur     = 4 * p*N_f * (l_pfe + b_pc + pi / 4 * (pi * (r_r - h_pc - h_ps) / p - b_pc))
        A_Cur     = k_fillr * h_pc        * 0.5 / N_f * (pi * (r_r - h_pc - h_ps)        / p - b_pc)
        A_Curcalc = k_fillr * h_pc * 1000 * 0.5 / N_f * (pi * (r_r - h_pc - h_ps) * 1000 / p - b_pc * 1000)
        Slot_Area = A_Cur * 2 * N_f / k_fillr # (not used)
        R_r = rho_Cu * l_Cur / A_Cur
        
        # field winding current density
        
        J_f = I_f / A_Curcalc
        
        # calculating air flux density
        
        B_gfm = mu_0 * N_f * I_f / (g_1 * (1 + k_s))  # No load air gap flux density
        
        B_g = B_gfm * 4*sin(0.5 * b_p * pi / tau_p) / pi  # fundamental component
        B_symax = tau_p * B_g / pi / h_ys                 # stator yoke flux density
        L_fg = 2 * mu_0 * p*l_s * 4*N_f**2 * ((h_ps / (tau_p - b_p)) + (h_pc / (3 * pi * (r_r - h_pc - h_ps) / p - b_pc))) #  (not used)
        
        # calculating no-load voltage and stator current
        
        E_s = 2 * N_s * l_s * rad_ag * k_wd * om_m * B_g / sqrt(2) # no-load voltage
        I_s = (E_s - (E_s**2 - 4 * R_s * machine_rating / m)**0.5) / (2 * R_s)
        
        # Calculating stator winding current density and specific current loading
        
        A_1 = 6 * N_s * I_s / (pi * dia)
        J_s = I_s / A_Cuscalc
        
        # Calculating magnetic loading in other parts of the machine
        
        delta_m = 0  # Initialising load angle
        
        # peak flux density in pole core, rotor yoke and stator teeth
        
        B_pc = (1 / b_pc) * ((2 * tau_p / pi) * B_g * cos(delta_m) + (2 * mu_0 * I_f * N_f * ((2 * h_ps / (tau_p - b_p)) + (h_pc / (tau_p - b_pc)))))
        B_rymax = 0.5 * b_pc * B_pc / h_yr
        B_tmax = (B_gfm + B_g) * tau_s * 0.5 / b_t
        
        # Calculating leakage inductances in the stator
        
        L_ssigmas = 2 * mu_0 * l_s * N_s**2 / p / q1 * ((h_s - h_w) / (3 * b_s) + h_w / b_so)  # slot leakage inductance
        L_ssigmaew =    mu_0 * 1.2 * N_s**2 / p * 1.2 * (2 / 3 * tau_p + 0.01)                 # end winding leakage inductance
        L_ssigmag = 2 * mu_0 * l_s * N_s**2 / p / q1 * (5 * (g / b_so) / (5 + 4 * (g / b_so))) # tooth tip leakage inductance
        L_ssigma = (L_ssigmas + L_ssigmaew + L_ssigmag)  # stator leakage inductance
        
        # Calculating effective air gap
        
        '''
        What is the source of this function that combines 1st and 13th powers? Very suspicious...
        Inputs appear to be in the range of 0.45 to 2.2, so outputs are 180 to 178000
        
        def airGapFn(B, fact):
            val = 400 * B + 7 * B**13
            ans = val * fact
            sys.stderr.write('aGF: B {} val {} ans {}\n'.format(B, val, ans))
            return val
        
        At_t = h_s * airGapFn(B_tmax, h_s)
        At_sy = tau_p * 0.5 * airGapFn(B_symax, tau_p/2)
        At_pc = (h_pc + h_ps) * airGapFn(B_pc, h_pc + h_ps)
        At_ry = tau_p * 0.5 * airGapFn(B_rymax, tau_p/2)
        '''
        At_g  = g_1 * B_gfm / mu_0
        At_t  = h_s           * (400 * B_tmax  + 7 * B_tmax**13)
        At_sy = tau_p * 0.5   * (400 * B_symax + 7 * B_symax**13)
        At_pc = (h_pc + h_ps) * (400 * B_pc    + 7 * B_pc**13)
        At_ry = tau_p * 0.5   * (400 * B_rymax + 7 * B_rymax**13)
        g_eff = (At_g + At_t + At_sy + At_pc + At_ry) * g_1 / At_g
        
        L_m = 6 * k_wd**2 * N_s**2 * mu_0 * rad_ag * l_s / pi / g_eff / p**2
        B_r1 = (mu_0 * I_f * N_f * 4 * sin(0.5 * (b_p / tau_p) * pi)) / g_eff / pi # (not used)
        
        # Calculating direct axis and quadrature axes inductances
        L_dm = (b_p / tau_p +(1 / pi) * sin(pi * b_p / tau_p)) * L_m
        L_qm = (b_p / tau_p -(1 / pi) * sin(pi * b_p / tau_p) + 2 / (3 * pi) * cos(b_p * pi / 2 * tau_p)) * L_m
        
        # Calculating actual load angle
        
        delta_m = (atan(om_e * L_qm * I_s / E_s))
        L_d = L_dm + L_ssigma # (not used)
        L_q = L_qm + L_ssigma # (not used)
        I_sd = I_s * sin(delta_m)
        I_sq = I_s * cos(delta_m)
        
        # induced voltage
        
        E_p = om_e * L_dm * I_sd + sqrt(E_s**2 - (om_e * L_qm * I_sq)**2) # (not used)
        # M_sf = mu_0 * 8*rad_ag * l_s * k_wd * N_s * N_f * sin(0.5 * b_p / tau_p * pi) / (p * g_eff * pi)
        # I_f1 = sqrt(2) * (E_p) / (om_e * M_sf)
        # I_f2 = (E_p / E_s) * B_g * g_eff * pi / (4 * N_f * mu_0 * sin(pi * b_p / 2/tau_p))
        # phi_max_stator = k_wd * N_s * pi * rad_ag * l_s * 2*mu_0 * N_f * I_f * 4*sin(0.5 * b_p / tau_p / pi) / (p * pi * g_eff * pi)
        # M_sf = mu_0 * 8*rad_ag * l_s * k_wd * N_s * N_f * sin(0.5 * b_p / tau_p / pi) / (p * g_eff * pi)
        
        L_tot = l_s + 2 * tau_p
        
        # Excitation power
        V_fn = 500
        Power_excitation = V_fn * 2*I_f   # total rated power in excitation winding
        Power_ratio = Power_excitation * 100 / machine_rating
        
        # Calculating Electromagnetically Active mass
        L_tot = l_s + 2 * tau_p # (not used)
        V_Cuss = m * l_Cus * A_Cus                                                     # volume of copper in stator
        V_Cusr = l_Cur * A_Cur                                                         # volume of copper in rotor
        V_Fest = (l_s * pi * ((rad_ag + h_s)**2 - rad_ag**2) - 2 * m*q1 * p*b_s * h_s * l_s) # volume of iron in stator tooth
        V_Fesy = l_s * pi * ((rad_ag + h_s + h_ys)**2 - (rad_ag + h_s)**2)                   # volume of iron in stator yoke
        V_Fert = 2 * p*l_pfe * (h_pc * b_pc + b_p * h_ps)                              # volume of iron in rotor pole
        V_Fery = l_pfe * pi * ((r_r - h_ps - h_pc)**2 - (r_r - h_ps - h_pc - h_yr)**2) # volume of iron in rotor yoke
        
        Copper = (V_Cuss + V_Cusr) * rho_Copper
        M_Fest = V_Fest * rho_Fe
        M_Fesy = V_Fesy * rho_Fe
        M_Fert = V_Fert * rho_Fe
        M_Fery = V_Fery * rho_Fe
        Iron = M_Fest + M_Fesy + M_Fert + M_Fery
        
        I_snom = machine_rating / (3 * E_s * cos_phi)
        
        ## Optional## Calculating mmf ratio
        F_1no_load = 3 * 2**0.5 * N_s * k_wd * I_s / (pi * p) # (not used)
        Nf_If_no_load = N_f * I_f
        F_1_rated = (3 * 2**0.5 * N_s * k_wd * I_srated) / (pi * p)
        Nf_If_rated = 2 * Nf_If_no_load
        Load_mmf_ratio = Nf_If_rated / F_1_rated
        
        ## Calculating losses
        #1. Copper losses
        K_R = 1.2
        P_Cuss = m * I_snom**2 * R_s * K_R
        P_Cusr = I_f**2 * R_r
        P_Cusnom_total = P_Cuss + P_Cusr
        
        #2. Iron losses ( Hysteresis and Eddy currents)
        P_Hyys = M_Fesy * (B_symax / 1.5)**2 * (P_Fe0h * om_e / (2 * pi * 60))      # Hysteresis losses in stator yoke
        P_Ftys = M_Fesy * (B_symax / 1.5)**2 * (P_Fe0e * (om_e / (2 * pi * 60))**2) # Eddy losses in stator yoke
        P_Fesynom = P_Hyys + P_Ftys
        P_Hyd = M_Fest * (B_tmax / 1.5)**2 * (P_Fe0h * om_e / (2 * pi * 60))        # Hysteresis losses in stator teeth
        P_Ftd = M_Fest * (B_tmax / 1.5)**2 * (P_Fe0e * (om_e / (2 * pi * 60))**2)   # Eddy losses in stator teeth
        P_Festnom = P_Hyd + P_Ftd
        
        # brushes
        delta_v = 1
        n_brushes = (I_f * 2 / 120)
        
        if (n_brushes<0.5):
            n_brushes = 1
        else:
            n_brushes = np.round(n_brushes)
            
        #3. brush losses
        
        p_b = 2 * delta_v * (I_f)
        Losses = P_Cusnom_total + P_Festnom + P_Fesynom + p_b
        gen_eff = machine_rating * 100 / (Losses + machine_rating)
        
        ################################################## Structural  Design ########################################################
        
        ## Structural deflection calculations
        
        # rotor structure
        
        q3          = B_g**2 / 2/mu_0           # normal component of Maxwell's stress
        #l           = l_s                       # l - stator core length - now using l_s everywhere
        l_b         = 2 * tau_p                 # end winding length # (not used)
        l_e         = l_s + 2 * 0.001 * rad_ag     # equivalent core length # (not used)
        a_r         = (b_r * d_r) - ((b_r - 2 * t_wr) * (d_r - 2 * t_wr))  # cross-sectional area of rotor armms
        A_r         = l_s * t                   # cross-sectional area of rotor cylinder
        N_r         = np.round(n_r)
        theta_r     = pi / N_r                  # half angle between spokes
        I_r         = l_s * t**3 / 12             # second moment of area of rotor cylinder
        I_arm_axi_r = ((b_r * d_r**3) - ((b_r - 2 * t_wr) * (d_r - 2 * t_wr)**3)) / 12  # second moment of area of rotor arm
        I_arm_tor_r = ((d_r * b_r**3) - ((d_r - 2 * t_wr) * (b_r - 2 * t_wr)**3)) / 12  # second moment of area of rotot arm w.r.t torsion
        R           = r_r - h_ps - h_pc - 0.5 * h_yr
        R_1         = R - h_yr * 0.5            # inner radius of rotor cylinder
        k_1         = sqrt(I_r / A_r)           # radius of gyration
        m1          = (k_1 / R)**2 
        c           = R / 500 # (not used)
        
        u_all_r     = R / 10000 # allowable radial deflection 
        b_all_r     = 2 * pi * R_o / N_r  # allowable circumferential arm dimension 
        
        # Calculating radial deflection of rotor structure according to Mc Donald's
        Numer = R**3 * ((0.25 * (sin(theta_r) - (theta_r * cos(theta_r))) / (sin(theta_r))**2) - (0.5 / sin(theta_r)) + (0.5 / theta_r))
        Pov   = ((theta_r / (sin(theta_r))**2) + 1 / tan(theta_r)) * ((0.25 * R / A_r) + (0.25 * R**3 / I_r))
        Qov   = R**3 / (2 * I_r * theta_r * (m1 + 1))
        Lov   = (R_1 - R_o) / a_r
        Denom = I_r * (Pov - Qov + Lov) # radial deflection % rotor
        u_Ar  = (q3 * R**2 / E / h_yr) * (1 + Numer / Denom)
        
        # Calculating axial deflection of rotor structure
        
        w_r         = rho_Fes * g1 * sin(phi) * a_r * N_r
        mass_st_lam = rho_Fe * 2*pi * (R + 0.5 * h_yr) * l_s * h_yr                             # mass of rotor yoke steel
        W           = g1 * sin(phi) * (mass_st_lam + (V_Cusr * rho_Copper) + M_Fert) / N_r  # weight of rotor cylinder
        l_ir        = R                                      # length of rotor arm beam at which rotor cylinder acts
        l_iir       = R_1
        
        y_Ar        = (W * l_ir**3 / 12 / E / I_arm_axi_r) + (w_r * l_iir**4 / 24 / E / I_arm_axi_r)  # axial deflection
        
        # Calculating torsional deflection of rotor structure
        
        z_all_r     = radians(0.05 * R)  # allowable torsional deflection
        z_A_r       = (2 * pi * (R - 0.5 * h_yr) * l_s / N_r) * sigma * (l_ir - 0.5 * h_yr)**3 / 3 / E / I_arm_tor_r # circumferential deflection
        
        # STATOR structure
        
        A_st        = l_s * t_s
        a_s         = (b_st * d_s) - ((b_st - 2 * t_ws) * (d_s - 2 * t_ws))
        N_st        = np.round(n_s)
        theta_s     = pi / N_st 
        I_st        = l_s * t_s**3 / 12
        I_arm_axi_s = ((b_st * d_s**3) - ((b_st - 2 * t_ws) * (d_s - 2 * t_ws)**3)) / 12  # second moment of area of stator arm
        I_arm_tor_s = ((d_s * b_st**3) - ((d_s - 2 * t_ws) * (b_st - 2 * t_ws)**3)) / 12  # second moment of area of rotot arm w.r.t torsion
        R_st        = rad_ag + h_s + h_ys * 0.5
        R_1s        = R_st - h_ys * 0.5
        k_2         = sqrt(I_st / A_st)
        m2          = (k_2 / R_st)**2
        
        # allowable deflections
        
        b_all_s   = 2 * pi * R_o / N_st
        u_all_s   = R_st / 10000
        y_all     = 2 * l_s / 100         # allowable axial     deflection
        z_all_s   = radians(0.05 * R_st)  # allowable torsional deflection
        
        # Calculating radial deflection according to McDonald's
        
        Numers = R_st**3 * ((0.25 * (sin(theta_s) - (theta_s * cos(theta_s))) / (sin(theta_s))**2) - (0.5 / sin(theta_s)) + (0.5 / theta_s))
        Povs   = ((theta_s / (sin(theta_s))**2) + 1 / tan(theta_s)) * ((0.25 * R_st / A_st) + (0.25 * R_st**3 / I_st))
        Qovs   = R_st**3 / (2 * I_st * theta_s * (m2 + 1))
        Lovs   = (R_1s - R_o) * 0.5 / a_s
        Denoms = I_st * (Povs - Qovs + Lovs)
        R_out  = (R / 0.995 + h_s + h_ys)
        u_As   = (q3 * R_st**2 / E / t_s) * (1 + Numers / Denoms)
        
        # Calculating axial deflection according to McDonald
        
        l_is          = R_st - R_o
        l_iis         = l_is
        l_iiis        = l_is                                                  # length of rotor arm beam at which self-weight acts
        mass_st_lam_s = M_Fest + pi * l_s * rho_Fe * ((R_st + 0.5 * h_ys)**2 - (R_st - 0.5 * h_ys)**2)
        W_is          = g1 * sin(phi) * (rho_Fes * l_s * d_s**2 * 0.5) # weight of rotor cylinder                              
        W_iis         = g1 * sin(phi) * (V_Cuss * rho_Copper + mass_st_lam_s) / 2/N_st
        w_s           = rho_Fes * g1 * sin(phi) * a_s * N_st
        
        X_comp1 = W_is  * l_is**3   / 12 / E / I_arm_axi_s
        X_comp2 = W_iis * l_iis**4  / 24 / E / I_arm_axi_s
        X_comp3 = w_s   * l_iiis**4 / 24 / E / I_arm_axi_s
        
        y_As    = X_comp1 + X_comp2 + X_comp3  # axial deflection
        
        # Calculating torsional deflection
        
        z_A_s  = 2 * pi * (R_st + 0.5 * t_s) * l_s / (2 * N_st) * sigma * (l_is + 0.5 * t_s)**3 / 3 / E / I_arm_tor_s 
        
        # tangential stress constraints
        
        TC1 = T / (2 * pi * sigma)
        TC2 = R**2 * l_s
        TC3 = R_st**2 * l_s
        
        mass_stru_steel  = 2 * (N_st * (R_1s - R_o) * a_s * rho_Fes)
        
        # Calculating inactive mass and total mass
        
        Structural_mass = mass_stru_steel + (N_r * (R_1 - R_o) * a_r * rho_Fes)
        
        Mass = Copper + Iron + Structural_mass
        
        I = np.zeros(3)
        # Calculating mass moments of inertia and center of mass
        I[0]   = (0.50 * Mass*R_out**2)
        I[1]   = (0.25 * Mass*R_out**2 + Mass * l_s**2 / 12) 
        I[2]   = I[1]
        cm = np.zeros(3)
        cm[0]  = shaft_cm[0] + shaft_length / 2. + l_s / 2
        cm[1]  = shaft_cm[1]
        cm[2]  = shaft_cm[2]

        outputs['B_symax']           = B_symax
        outputs['B_tmax']            = B_tmax
        outputs['B_rymax']           = B_rymax
        outputs['B_gfm']             = B_gfm
        outputs['B_g']               = B_g
        outputs['B_pc']              = B_pc
        outputs['N_s']               = N_s
        outputs['b_s']               = b_s

        outputs['b_t']               = b_t
        outputs['A_Cuscalc']         = A_Cuscalc
        outputs['A_Curcalc']         = A_Curcalc
        outputs['b_p']               = b_p
        outputs['h_p']               = h_p
        outputs['p']                 = p
        outputs['E_s']               = E_s
        outputs['f']                 = f

        outputs['I_s']               = I_s
        outputs['R_s']               = R_s
        outputs['L_m']               = L_m
        outputs['A_1']               = A_1
        outputs['J_s']               = J_s
        outputs['R_r']               = R_r
        outputs['Losses']            = Losses

        outputs['Load_mmf_ratio']    = Load_mmf_ratio
        outputs['Power_ratio']       = Power_ratio
        outputs['n_brushes']         = n_brushes
        outputs['J_f']               = J_f
        outputs['K_rad']             = K_rad
        outputs['gen_eff']           = gen_eff
        outputs['S']                 = S

        outputs['Slot_aspect_ratio'] = Slot_aspect_ratio
        outputs['Copper']            = Copper
        outputs['Iron']              = Iron
        outputs['u_Ar']              = u_Ar
        outputs['y_Ar']              = y_Ar

        outputs['z_A_r']             = z_A_r
        outputs['u_As']              = u_As
        outputs['y_As']              = y_As
        outputs['z_A_s']             = z_A_s
        outputs['u_all_r']           = u_all_r
        outputs['u_all_s']           = u_all_s

        outputs['y_all']             = y_all
        outputs['z_all_s']           = z_all_s
        outputs['z_all_r']           = z_all_r
        outputs['b_all_s']           = b_all_s
        outputs['b_all_r']           = b_all_r
        outputs['TC1']               = TC1

        outputs['TC2']               = TC2
        outputs['TC3']               = TC3
        outputs['R_out']             = R_out
        outputs['Structural_mass']   = Structural_mass
        outputs['Mass']              = Mass
        outputs['cm']                = cm
        outputs['I']                 = I
        
        



"""
Demonstration of setting up an OpenMDAO 1.x problem using the FST8Workflow component
(in FST8_aeroelasticsolver), which executes the FST8 reader, writer, and wrapper and assigns
all variables in the FAST outlist to OpenMDAO 'Unknowns'. It also
implements an "input config" function which allows the user to put all variables that they
wish to explicitly define into a dictionary. The input config function assigns these
variables to the correct locations in the variable tree.
"""
# Hacky way of doing relative imports
import numpy as np
import os, sys
sys.path.insert(0, os.path.abspath(".."))
import matplotlib.pyplot as plt
from openmdao.api import Group, Problem, Component, IndepVarComp, ParallelGroup
from openmdao.api import SqliteRecorder
from wisdem.aeroelasticse.FAST_mdao.FST8_aeroelasticsolver import FST8Workflow
from wisdem.aeroelasticse.Turbsim_mdao.turbsim_openmdao import turbsimGroup

# Initial OpenMDAO problem setup
top = Problem()
root = top.root = Group()

# Setup input config--file/directory locations, executable, types
caseid = "omdaoCase1.fst"
config = {}
config['fst_masterfile'] = 'Test01.fst' 
config['fst_masterdir']= '../FAST_mdao/wrapper_examples/FST8inputfiles'
config['fst_runfile'] = caseid
config['fst_rundir'] = './rundir/'
config['fst_exe'] = 'openfast'
#config['fst_exe'] = '../../../../../FAST_v8/bin/FAST_glin64'
#config['libmap'] = '../../../../../FAST_v8/bin/libmap-1.20.10.dylib'
config['ad_file_type'] = 1 

# Additional parameters
TMAX = 90
config['TMax'] = TMAX

# Add Turbsim then FAST
root.add('fast_component', FST8Workflow(config, caseid))
root.add('turbsim_component', turbsimGroup())

# # Set up recorder
recorder = SqliteRecorder('omdaoCase1.sqlite')
top.driver.add_recorder(recorder)

# Perform setup and run OpenMDAO problem
top.setup()

top.root.turbsim_component.wrapper.turbsim_exe = '/Users/jquick/TurbSim/bin/TurbSim_glin64'
#top.root.fast_component.writer.fst_vt.steady_wind_params.HWindSpeed = 15.12345
top.root.fast_component.writer.fst_vt.turbsim_wind_params.Filename = './turbsim_default.bts' # one directory below true location....
top.root.fast_component.writer.fst_vt.inflow_wind.WindType = 3
top.root.fast_component.writer.fst_vt.fst_sim_ctrl.TMax = TMAX
#top.root.turbsim_component.run_dir = './rundir/'
#for i in range(2): # Can't set fst rundir
#top.root.turbsim_component.run_dir = './rundir%i/'%i
top.root.turbsim_component.run_dir = './rundir'
#top.root.fast_component.writer.fst_file = './rundir%i/'%i
top.root.turbsim_component.writer.turbsim_vt.runtime_options.RandSeed1 = 10000 + 11
top.root.turbsim_component.execute()
top.run()
plt.plot(top['fast_component.RootMxc1'])
np.save('fout', top['fast_component.RootMxc1'])
top.cleanup()   #Good practice, especially when using recorder
plt.savefig('seeds.pdf')



#-------------------------------------------------------------------------------
# Name:        GetSheetByName.py
# Purpose:     This module contains a function to find index of sheet in excel file that corresponds to a given name
#
# Author:      rdamiani
#
# Created:     6/01/2014 - Based on stackoverflow
# Copyright:   (c) rdamiani 2014
# Licence:     <Apache>
#-------------------------------------------------------------------------------
import itertools
#______________________________________________________________________________#
def get_sheet_idx(book, name):
    """Get a sheet by name from xlwt.Workbook, a strangely missing method.
    Returns None if no sheet with the given name is present.
    """
    # Note, we have to use exceptions for flow control because the
    # xlwt API is broken and gives us no other choice.
    try:
        for idx in itertools.count():
            sheet = book.get_sheet(idx)
            if sheet.name == name:
                return idx
    except IndexError:
        return None
#______________________________________________________________________________#

if __name__ == '__main__':

    get_sheet_idx('test.xls','testsheet')




from __future__ import print_function
import numpy as np
from pprint import pprint
from openmdao.api import IndepVarComp, ExplicitComponent, Group, Problem, ScipyOptimizeDriver, SqliteRecorder, NonlinearRunOnce, DirectSolver
try:
    from openmdao.api import pyOptSparseDriver
except:
    pass
from wisdem.rotorse.rotor import RotorSE, Init_RotorSE_wRefBlade
from wisdem.rotorse.rotor_geometry_yaml import ReferenceBlade
from wisdem.towerse.tower import TowerSE
from wisdem.commonse import NFREQ
from wisdem.commonse.environment import PowerWind, LogWind
from wisdem.commonse.turbine_constraints import TurbineConstraints
from wisdem.turbine_costsse.turbine_costsse_2015 import Turbine_CostsSE_2015
from wisdem.orbit.api.wisdem.fixed import Orbit
from wisdem.plant_financese.plant_finance import PlantFinance
from wisdem.drivetrainse.drivese_omdao import DriveSE

from wisdem.commonse.mpi_tools import MPI

# np.seterr(all ='raise')
        
# Group to link the openmdao components
class MonopileTurbine(Group):

    def initialize(self):
        self.options.declare('RefBlade')
        self.options.declare('FASTpref', default={})
        self.options.declare('Nsection_Tow', default = 6)
        self.options.declare('VerbosityCosts', default = True)
        self.options.declare('user_update_routine',     default=None)
        
        
    def setup(self):
        
        RefBlade            = self.options['RefBlade']
        Nsection_Tow        = self.options['Nsection_Tow']        
        user_update_routine = self.options['user_update_routine']
        if 'Analysis_Level' in self.options['FASTpref']:
            Analysis_Level  = self.options['FASTpref']['Analysis_Level']
        else:
            Analysis_Level  = 0
        
        # Define all input variables from all models
        myIndeps = IndepVarComp()
        myIndeps.add_discrete_output('crane',    False)

        # Turbine Costs
        myIndeps.add_discrete_output('bearing_number', 0)
        
        # Tower and Frame3DD options
        myIndeps.add_output('project_lifetime',             0.0, units='yr')
        myIndeps.add_output('max_taper_ratio',              0.0)
        myIndeps.add_output('min_diameter_thickness_ratio', 0.0)

        # Environment
        myIndeps.add_output('wind_bottom_height',   0.0, units='m')
        myIndeps.add_output('wind_beta',            0.0, units='deg')
        myIndeps.add_output('cd_usr', -1.)


        # Environment Offshore
        myIndeps.add_output('offshore',           True)
        myIndeps.add_output('water_depth',        0.0, units='m')

        # Design standards
        myIndeps.add_output('gamma_b', 0.0)
        myIndeps.add_output('gamma_n', 0.0)

        # Column
        myIndeps.add_output('morison_mass_coefficient', 2.0)
        myIndeps.add_output('material_density',         0.0, units='kg/m**3')
        myIndeps.add_output('E',                        0.0, units='N/m**2')
        myIndeps.add_output('yield_stress',             0.0, units='N/m**2')

        # Pontoons
        myIndeps.add_output('G', 0.0, units='N/m**2')
        
        # LCOE
        myIndeps.add_output('labor_cost_rate',      0.0, units='USD/min')
        myIndeps.add_output('material_cost_rate',   0.0, units='USD/kg')
        myIndeps.add_output('painting_cost_rate',   0.0, units='USD/m**2')
        myIndeps.add_discrete_output('number_of_turbines', 0)
        myIndeps.add_output('annual_opex',          0.0, units='USD/kW/yr') # TODO: Replace with output connection
        myIndeps.add_output('fixed_charge_rate',    0.0)
        myIndeps.add_output('wake_loss_factor',     0.0)
        
        self.add_subsystem('myIndeps', myIndeps, promotes=['*'])

        
        # Add components
        self.add_subsystem('rotorse', RotorSE(RefBlade=RefBlade,
                                              npts_coarse_power_curve=20,
                                              npts_spline_power_curve=200,
                                              regulation_reg_II5=True,
                                              regulation_reg_III=True,
                                              Analysis_Level=Analysis_Level,
                                              FASTpref=self.options['FASTpref'],
                                              topLevelFlag=True,
                                              user_update_routine=user_update_routine), promotes=['*'])
        
        self.add_subsystem('drive', DriveSE(debug=False,
                                            number_of_main_bearings=1,
                                            topLevelFlag=False),
                           promotes=['machine_rating', 'overhang',
                                     'hub_mass','bedplate_mass','gearbox_mass','generator_mass','hss_mass','hvac_mass','lss_mass','cover_mass',
                                     'pitch_system_mass','platforms_mass','spinner_mass','transformer_mass','vs_electronics_mass','yaw_mass'])
        
        # Tower and substructure
        self.add_subsystem('tow',TowerSE(nLC=1,
                                         nPoints=Nsection_Tow+1,
                                         nFull=3*Nsection_Tow+1,
                                         wind='PowerWind',
                                         topLevelFlag=False,
                                         monopile=True),
                           promotes=['water_density','water_viscosity','wave_beta',
                                     'significant_wave_height','significant_wave_period',
                                     'material_density','E','G','tower_section_height',
                                     'tower_wall_thickness', 'tower_outer_diameter',
                                     'tower_outfitting_factor','tower_buckling_length',
                                     'transition_piece_mass','transition_piece_height',
                                     'max_taper','min_d_to_t','rna_mass','rna_cg','rna_I',
                                     'tower_add_gravity','tower_mass','tower_I_base','hub_height',
                                     'foundation_height','soil_G','soil_nu','suctionpile_depth','suctionpile_depth_diam_ratio',
                                     'monopile_mass','monopile_cost','monopile_length','tower_raw_cost',
                                     'gamma_f','gamma_m','gamma_b','gamma_n','gamma_fatigue',
                                     'labor_cost_rate','material_cost_rate','painting_cost_rate','z_full','d_full','t_full',
                                     'DC','shear','geom','tower_force_discretization','nM','Mmethod','lump','tol','shift'])

        # Turbine constraints
        self.add_subsystem('tcons', TurbineConstraints(nFull=3*(Nsection_Tow+1)+1), promotes=['*'])
        
        # Turbine costs
        self.add_subsystem('tcost', Turbine_CostsSE_2015(verbosity=self.options['VerbosityCosts'], topLevelFlag=False), promotes=['*'])

        # Use ORBIT for BOS costs
        self.add_subsystem('orbit', Orbit(), promotes=['wtiv', 'feeder', 'num_feeders', 'oss_install_vessel',
                                                       'hub_height',
                                                       'number_of_turbines',
                                                       'tower_mass',
                                                       'monopile_length',
                                                       'monopile_mass',
                                                       'transition_piece_mass',
                                                       'site_distance',
                                                       'site_distance_to_landfall',
                                                       'site_distance_to_interconnection',
                                                       'plant_turbine_spacing',
                                                       'plant_row_spacing',
                                                       'plant_substation_distance',
                                                       'tower_deck_space',
                                                       'nacelle_deck_space',
                                                       'blade_deck_space',
                                                       'port_cost_per_month',
                                                       'monopile_deck_space',
                                                       'transition_piece_deck_space',
                                                       #'scour_protection_depth',
                                                       'commissioning_pct',
                                                       'decommissioning_pct'])
        
        # LCOE Calculation
        self.add_subsystem('plantfinancese', PlantFinance(verbosity=self.options['VerbosityCosts']),
                           promotes=['machine_rating','lcoe', 'fixed_charge_rate','wake_loss_factor'])
        
    
        # Set up connections

        # Connections to DriveSE
        self.connect('nBlades',         'drive.number_of_blades')
        self.connect('diameter',        'drive.rotor_diameter')     
        self.connect('rated_Q',         'drive.rotor_torque')
        self.connect('rated_Omega',     'drive.rotor_rpm')
        self.connect('Fxyz_total',      'drive.Fxyz')
        self.connect('Mxyz_total',      'drive.Mxyz')
        self.connect('I_all_blades',        'drive.blades_I')
        self.connect('mass_one_blade',  'drive.blade_mass')
        self.connect('chord',           'drive.blade_root_diameter', src_indices=[0])
        self.connect('Rtip',            'drive.blade_length', src_indices=[0])
        self.connect('drivetrainEff',   'drive.drivetrain_efficiency', src_indices=[0])
        self.connect('tower_outer_diameter', 'drive.tower_top_diameter', src_indices=[-1])

        self.connect('material_density', 'tow.tower.rho')

        
        # Connections to TowerSE
        self.connect('drive.top_F',         'tow.pre.rna_F')
        self.connect('drive.top_M',         'tow.pre.rna_M')
        self.connect('drive.rna_I_TT',            ['rna_I','tow.pre.mI'])
        self.connect('drive.rna_cm',              ['rna_cg','tow.pre.mrho'])
        self.connect('drive.rna_mass',            ['rna_mass','tow.pre.mass'])
        self.connect('rs.gust.V_gust',          'tow.wind.Uref')
        self.connect('wind_reference_height',   ['tow.wind.zref','wind.zref'])
        # self.connect('wind_bottom_height',      ['tow.wind.z0','tow.wave.z_surface', 'wind.z0'])  # offshore
        self.connect('wind_bottom_height',      ['tow.wind.z0', 'wind.z0'])
        self.connect('shearExp',                ['tow.wind.shearExp'])
        # self.connect('morison_mass_coefficient','tow.cm')                                         # offshore
        self.connect('yield_stress',            'tow.sigma_y')
        self.connect('max_taper_ratio',         'max_taper')
        self.connect('min_diameter_thickness_ratio', 'min_d_to_t')

        self.connect('rho',         'tow.windLoads.rho')
        self.connect('mu',          'tow.windLoads.mu')
        self.connect('wind_beta',   'tow.windLoads.beta')

        self.connect('nBlades',                 'blade_number')
        self.connect('control_maxOmega',        'rotor_omega')
        self.connect('tow.post.structural_frequencies', 'tower_freq')        
                
        # Connections to TurbineCostSE
        self.connect('mass_one_blade',              ['blade_mass','orbit.blade_mass'])
        self.connect('drive.mainBearing.mb_mass',   'main_bearing_mass')
        self.connect('total_blade_cost',            'blade_cost_external')
        self.connect('tower_raw_cost',              'tower_cost_external')

        # Connections to ORBIT
        self.connect('water_depth', 'orbit.site_depth')
        self.connect('rs.gust.V_gust', 'orbit.site_mean_windspeed')
        self.connect('machine_rating', 'orbit.turbine_rating')
        self.connect('rated_V', 'orbit.turbine_rated_windspeed')
        self.connect('turbine_cost_kW', 'orbit.turbine_capex')
        self.connect('diameter', 'orbit.turbine_rotor_diameter')
        self.connect('drive.nacelle_mass', 'orbit.nacelle_mass')
        self.connect('tower_outer_diameter', 'orbit.monopile_diameter', src_indices=[0])
        
        # Connections to PlantFinanceSE
        self.connect('AEP',                 'plantfinancese.turbine_aep')
        self.connect('turbine_cost_kW',     'plantfinancese.tcc_per_kW')
        self.connect('number_of_turbines',  'plantfinancese.turbine_number')
        self.connect('orbit.total_capex_kW','plantfinancese.bos_per_kW')
        self.connect('annual_opex',         'plantfinancese.opex_per_kW')
    

def Init_MonopileTurbine(prob, blade, Nsection_Tow, Analysis_Level = 0, fst_vt = {}):

    prob = Init_RotorSE_wRefBlade(prob, blade, Analysis_Level = Analysis_Level, fst_vt = fst_vt)

    # Environmental parameters for the tower
    # prob['wind_reference_speed']           = 11.0
    prob['wind_reference_height']          = prob['hub_height']
    prob['shearExp']                       = 0.11
    prob['rho']                            = 1.225
    prob['mu']                             = 1.7934e-5
    prob['water_density']                  = 1025.0
    prob['water_viscosity']                = 1.3351e-3
    prob['wind_beta'] = prob['wave_beta'] = 0.0
    prob['significant_wave_height']        = 5.0
    prob['significant_wave_period']        = 10.0

    # Steel properties for the tower
    prob['material_density']               = 7850.0
    prob['E']                              = 210e9
    prob['G']                              = 79.3e9
    prob['yield_stress']                   = 345e6
    prob['soil_G']                         = 140e6
    prob['soil_nu']                        = 0.4

    # Design constraints
    prob['max_taper_ratio']                = 0.4
    prob['min_diameter_thickness_ratio']   = 120.0

    # Safety factors
    prob['gamma_fatigue']   = 1.755 # (Float): safety factor for fatigue
    prob['gamma_f']         = 1.35  # (Float): safety factor for loads/stresses
    prob['gamma_m']         = 1.3   # (Float): safety factor for materials
    prob['gamma_freq']      = 1.1   # (Float): safety factor for resonant frequencies
    prob['gamma_n']         = 1.0
    prob['gamma_b']         = 1.1
    
    # Tower
    # prob['foundation_height']              = -prob['water_depth']
    # prob['tower_outer_diameter']           = np.linspace(10.0, 3.87, Nsection_Tow+1)
    prob['tower_outer_diameter']           = np.linspace(6.0, 3.87, Nsection_Tow+1)
    prob['tower_section_height']           = (prob['hub_height'] - prob['foundation_height']) / Nsection_Tow * np.ones(Nsection_Tow)
    prob['tower_wall_thickness']           = np.linspace(0.027, 0.019, Nsection_Tow)
    prob['tower_buckling_length']          = 30.0
    prob['tower_outfitting_factor']        = 1.07
    prob['transition_piece_mass']          = 100e3
    prob['transition_piece_height']        = 10.0

    prob['DC']      = 80.0
    prob['shear']   = True
    prob['geom']    = True
    prob['tower_force_discretization'] = 5.0
    prob['nM']      = 2
    prob['Mmethod'] = 1
    prob['lump']    = 0
    prob['tol']     = 1e-9
    prob['shift']   = 0.0
    
    # Plant size
    prob['project_lifetime'] = prob['lifetime'] = 20.0    
    prob['number_of_turbines']             = 200. * 1.e+006 / prob['machine_rating']
    prob['annual_opex']                    = 43.56 # $/kW/yr
    
    # For RNA
    prob['tower_add_gravity'] = True # Don't double count

    # For turbine costs
    prob['offshore']             = True
    prob['crane']                = False
    prob['bearing_number']       = 2
    prob['crane_cost']           = 0.0
    prob['labor_cost_rate']      = 3.0
    prob['material_cost_rate']   = 2.0
    prob['painting_cost_rate']   = 28.8
    
    # Gearbox
    prob['drive.gear_ratio']        = 96.76  # 97:1 as listed in the 5 MW reference document
    prob['drive.shaft_angle']       = prob['tilt']*np.pi / 180.0  # rad
    prob['drive.shaft_ratio']       = 0.10
    prob['drive.planet_numbers']    = [3, 3, 1]
    prob['drive.shrink_disc_mass']  = 333.3 * prob['machine_rating'] / 1e6  # estimated
    prob['drive.carrier_mass']      = 8000.0  # estimated
    prob['drive.flange_length']     = 0.5
    prob['overhang']                = 5.0
    prob['drive.distance_hub2mb']   = 1.912  # length from hub center to main bearing, leave zero if unknown
    prob['drive.gearbox_input_xcm'] = 0.1
    prob['drive.hss_input_length']  = 1.5
    prob['drive.yaw_motors_number'] = 1
    
    return prob
    

if __name__ == "__main__":
    
    optFlag = True



    # Reference rotor design
    fname_schema          = "../../rotorse/turbine_inputs/IEAontology_schema.yaml"
    fname_input           = "../../rotorse/turbine_inputs/nrel5mw_mod_update.yaml"
    Analysis_Level        = 0 # 0: Run CCBlade; 1: Update FAST model at each iteration but do not run; 2: Run FAST w/ ElastoDyn; 3: (Not implemented) Run FAST w/ BeamDyn
    # Initialize blade design
    refBlade = ReferenceBlade()
    refBlade.verbose      = True
    refBlade.NINPUT       = 8
    refBlade.NPTS         = 50
    refBlade.spar_var     = ['Spar_Cap_SS', 'Spar_Cap_PS'] # SS, then PS
    refBlade.te_var       = 'TE_reinforcement'
    refBlade.validate     = False
    refBlade.fname_schema = fname_schema
    blade = refBlade.initialize(fname_input)
    # Initialize tower design
    Nsection_Tow = 6

    
    # Initialize OpenMDAO problem and FloatingSE Group
    if MPI:
        num_par_fd = MPI.COMM_WORLD.Get_size()
        prob = Problem(model=Group(num_par_fd=num_par_fd))
        prob.model.approx_totals(method='fd')
        prob.model.add_subsystem('comp', MonopileTurbine(RefBlade=blade, Nsection_Tow = Nsection_Tow, VerbosityCosts = True), promotes=['*'])
    else:
        prob = Problem()
        prob.model=MonopileTurbine(RefBlade=blade, Nsection_Tow = Nsection_Tow, VerbosityCosts = True)
    
    if optFlag:
        # --- Solver ---
        prob.driver  = ScipyOptimizeDriver()
        prob.driver.options['optimizer'] = 'SLSQP'
        prob.driver.options['tol']       = 1.e-6
        prob.driver.options['maxiter']   = 100

        #prob.driver = pyOptSparseDriver()
        #prob.driver.options['optimizer'] = 'CONMIN'
        # prob.driver.options['optimizer'] = 'SNOPT'
        # prob.driver.options['gradient method'] = "pyopt_fd"
        # ----------------------

        # --- Objective ---
        prob.model.add_objective('lcoe')
        # ----------------------

        # --- Design Variables ---
        indices_no_root         = range(2,refBlade.NINPUT)
        indices_no_root_no_tip  = range(2,refBlade.NINPUT-1)
        prob.model.add_design_var('chord_in',    indices = indices_no_root_no_tip, lower=0.5,      upper=7.0)
        prob.model.add_design_var('theta_in',    indices = indices_no_root,        lower=-5.0,     upper=20.0)
        prob.model.add_design_var('sparT_in',    indices = indices_no_root_no_tip, lower=0.001,    upper=0.200)
        prob.model.add_design_var('control_tsr',                                   lower=6.000,    upper=11.00)
        prob.model.add_design_var('tower_section_height', lower=5.0,  upper=80.0)
        prob.model.add_design_var('tower_outer_diameter', lower=3.87, upper=30.0)
        prob.model.add_design_var('tower_wall_thickness', lower=4e-3, upper=2e-1)
        # ----------------------
        
        # --- Constraints ---
        # Rotor
        prob.model.add_constraint('tip_deflection_ratio',                      upper=1.0)  
        # Tower
        prob.model.add_constraint('tow.height_constraint',         lower=-1e-2,upper=1.e-2)
        prob.model.add_constraint('tow.post.stress',                           upper=1.0)
        prob.model.add_constraint('tow.post.global_buckling',                  upper=1.0)
        prob.model.add_constraint('tow.post.shell_buckling',                   upper=1.0)
        prob.model.add_constraint('tow.weldability',                           upper=0.0)
        prob.model.add_constraint('tow.manufacturability',         lower=0.0)
        # prob.model.add_constraint('frequency1P_margin_low',              upper=1.0)
        # prob.model.add_constraint('frequency1P_margin_high', lower=1.0)
        # prob.model.add_constraint('frequencyNP_margin_low',              upper=1.0)
        # prob.model.add_constraint('frequencyNP_margin_high', lower=1.0)
        prob.model.add_constraint('frequencyNP_margin', upper=0.)
        prob.model.add_constraint('frequency1P_margin', upper=0.)
        prob.model.add_constraint('ground_clearance',        lower=20.0)
        # ----------------------
        
        # --- Recorder ---
        prob.driver.add_recorder(SqliteRecorder('log_opt.sql'))
        prob.driver.recording_options['includes'] = ['AEP','rc.total_blade_cost','lcoe','tip_deflection_ratio']
        prob.driver.recording_options['record_objectives']  = True
        prob.driver.recording_options['record_constraints'] = True
        prob.driver.recording_options['record_desvars']     = True
        # ----------------------


    prob.setup(check=True)
    
    prob = Init_MonopileTurbine(prob, blade, Nsection_Tow)
    prob.model.nonlinear_solver = NonlinearRunOnce()
    prob.model.linear_solver = DirectSolver()

    if not MPI:
        prob.model.approx_totals()

    prob.run_model()
    # prob.model.list_inputs(units=True)
    # prob.model.list_outputs(units=True)





    



from openmdao.api import Group, Component, IndepVarComp, DirectSolver, ScipyGMRES, Newton, NLGaussSeidel, Brent, RunOnce
from floatingse.floating import FloatingSE
from offshorebos.wind_obos_component import WindOBOS
from plant_financese.plant_finance import PlantFinance
from rotorse.rotor import RotorSE
from commonse.rna import RNA
from commonse.turbine_constraints import TurbineConstraints
from turbine_costsse.turbine_costsse_2015 import Turbine_CostsSE_2015

import numpy as np

        
class FloatingTurbine(Group):

    def __init__(self, RefBlade):
        super(FloatingTurbine, self).__init__()

        self.add('hub_height', IndepVarComp('hub_height', 0.0), promotes=['*'])

        # TODO: 
        #Weibull/Rayleigh CDF
        
        # Rotor
        self.add('rotor', RotorSE(RefBlade), promotes=['*'])

        # RNA
        self.add('rna', RNA(1), promotes=['downwind'])
        
        # Tower and substructure
        myfloat = FloatingSE()
        self.add('sm', myfloat, promotes=['*'])

        # Turbine constraints
        self.add('tcons', TurbineConstraints(myfloat.nFullTow), promotes=['*'])
        
        # Turbine costs
        self.add('tcost', Turbine_CostsSE_2015(), promotes=['*'])
        
        # Balance of station
        self.add('wobos', WindOBOS(), promotes=['*'])

        # LCOE Calculation
        self.add('lcoe', PlantFinance(), promotes=['*'])

        # Define all input variables from all models
        self.add('offshore',             IndepVarComp('offshore', True, pass_by_obj=True), promotes=['*'])
        self.add('crane',                IndepVarComp('crane', False, pass_by_obj=True), promotes=['*'])

        # Turbine Costs
        # REMOVE ONCE DRIVESE AND GENERATORSE ARE CONNECTED
        self.add('bearing_number',           IndepVarComp('bearing_number', 0, pass_by_obj=True), promotes=['*'])
        self.add('bedplate_mass',            IndepVarComp('bedplate_mass', 0.0), promotes=['*'])
        self.add('crane_cost',               IndepVarComp('crane_cost', 0.0), promotes=['*'])
        self.add('gearbox_mass',             IndepVarComp('gearbox_mass', 0.0), promotes=['*'])
        self.add('generator_mass',           IndepVarComp('generator_mass', 0.0), promotes=['*'])
        self.add('hss_mass',     IndepVarComp('hss_mass', 0.0), promotes=['*'])
        self.add('hvac_mass',   IndepVarComp('hvac_mass', 0.0), promotes=['*'])
        self.add('lss_mass',     IndepVarComp('lss_mass', 0.0), promotes=['*'])
        self.add('main_bearing_mass',        IndepVarComp('main_bearing_mass', 0.0), promotes=['*'])
        self.add('cover_mass',       IndepVarComp('cover_mass', 0.0), promotes=['*'])
        self.add('platforms_mass',   IndepVarComp('platforms_mass', 0.0), promotes=['*'])
        self.add('pitch_system_mass',        IndepVarComp('pitch_system_mass', 0.0), promotes=['*'])
        self.add('spinner_mass',             IndepVarComp('spinner_mass', 0.0), promotes=['*'])
        self.add('transformer_mass',         IndepVarComp('transformer_mass', 0.0), promotes=['*'])
        self.add('vs_electronics_mass', IndepVarComp('vs_electronics_mass', 0.0), promotes=['*'])
        self.add('yaw_mass',          IndepVarComp('yaw_mass', 0.0), promotes=['*'])
        
        # Tower
        #self.add('stress_standard_value',          IndepVarComp('stress_standard_value', 0.0), promotes=['*'])
        #self.add('fatigue_parameters',             IndepVarComp('fatigue_parameters', np.zeros(nDEL)), promotes=['*'])
        #self.add('fatigue_z',                      IndepVarComp('fatigue_z', np.zeros(nDEL)), promotes=['*'])
        #self.add('frame3dd_matrix_method',         IndepVarComp('frame3dd_matrix_method', 0, pass_by_obj=True), promotes=['*'])
        #self.add('compute_stiffnes',               IndepVarComp('compute_stiffnes', False, pass_by_obj=True), promotes=['*'])
        self.add('project_lifetime',               IndepVarComp('project_lifetime', 0.0), promotes=['*'])
        #self.add('lumped_mass_matrix',             IndepVarComp('lumped_mass_matrix', 0, pass_by_obj=True), promotes=['*'])
        #self.add('slope_SN',                       IndepVarComp('slope_SN', 0, pass_by_obj=True), promotes=['*'])
        #self.add('number_of_modes',                IndepVarComp('number_of_modes', 0, pass_by_obj=True), promotes=['*'])
        #self.add('compute_shear',                  IndepVarComp('compute_shear', True, pass_by_obj=True), promotes=['*'])
        #self.add('shift_value',                    IndepVarComp('shift_value', 0.0), promotes=['*'])
        #self.add('frame3dd_convergence_tolerance', IndepVarComp('frame3dd_convergence_tolerance', 1e-9), promotes=['*'])

        # TODO: Multiple load cases
        # Environment
        self.add('air_density',                IndepVarComp('air_density', 0.0), promotes=['*'])
        self.add('air_viscosity',              IndepVarComp('air_viscosity', 0.0), promotes=['*'])
        self.add('wind_reference_speed',       IndepVarComp('wind_reference_speed', 0.0), promotes=['*'])
        self.add('wind_reference_height',      IndepVarComp('wind_reference_height', 0.0), promotes=['*'])
        self.add('shearExp',                      IndepVarComp('shearExp', 0.0), promotes=['*'])
        self.add('wind_bottom_height',         IndepVarComp('wind_bottom_height', 0.0), promotes=['*'])
        self.add('wind_beta',                  IndepVarComp('wind_beta', 0.0), promotes=['*'])
        self.add('cd_usr',                     IndepVarComp('cd_usr', np.inf), promotes=['*'])

        # Environment
        self.add('water_depth',                IndepVarComp('water_depth', 0.0), promotes=['*'])
        self.add('water_density',              IndepVarComp('water_density', 0.0), promotes=['*'])
        self.add('water_viscosity',            IndepVarComp('water_viscosity', 0.0), promotes=['*'])
        self.add('wave_height',                IndepVarComp('wave_height', 0.0), promotes=['*'])
        self.add('wave_period',                IndepVarComp('wave_period', 0.0), promotes=['*'])
        self.add('mean_current_speed',          IndepVarComp('mean_current_speed', 0.0), promotes=['*'])
        #self.add('wave_beta',                  IndepVarComp('wave_beta', 0.0), promotes=['*'])
        #self.add('wave_velocity_z0',           IndepVarComp('wave_velocity_z0', 0.0), promotes=['*'])
        #self.add('wave_acceleration_z0',       IndepVarComp('wave_acceleration_z0', 0.0), promotes=['*'])
        
        # Design standards
        self.add('gamma_freq',      IndepVarComp('gamma_freq', 0.0), promotes=['*'])
        self.add('gamma_f',         IndepVarComp('gamma_f', 0.0), promotes=['*'])
        self.add('gamma_m',         IndepVarComp('gamma_m', 0.0), promotes=['*'])
        self.add('gamma_b',         IndepVarComp('gamma_b', 0.0), promotes=['*'])
        self.add('gamma_fatigue',   IndepVarComp('gamma_fatigue', 0.0), promotes=['*'])
        self.add('gamma_n',         IndepVarComp('gamma_n', 0.0), promotes=['*'])

        # RNA
        self.add('dummy_mass',                 IndepVarComp('dummy_mass', 0.0), promotes=['*'])
        self.add('hub_mass',                   IndepVarComp('hub_mass', 0.0), promotes=['*'])
        self.add('nac_mass',                   IndepVarComp('nac_mass', 0.0), promotes=['*'])
        self.add('hub_cm',                     IndepVarComp('hub_cm', np.zeros((3,))), promotes=['*'])
        self.add('nac_cm',                     IndepVarComp('nac_cm', np.zeros((3,))), promotes=['*'])
        self.add('hub_I',                      IndepVarComp('hub_I', np.zeros(6)), promotes=['*'])
        self.add('nac_I',                      IndepVarComp('nac_I', np.zeros(6)), promotes=['*'])
        self.add('rna_weightM',                IndepVarComp('rna_weightM', True, pass_by_obj=True), promotes=['*'])
        
        # Column
        self.add('morison_mass_coefficient',   IndepVarComp('morison_mass_coefficient', 0.0), promotes=['*'])
        self.add('material_density',           IndepVarComp('material_density', 0.0), promotes=['*'])
        self.add('E',                          IndepVarComp('E', 0.0), promotes=['*'])
        self.add('nu',                         IndepVarComp('nu', 0.0), promotes=['*'])
        self.add('yield_stress',               IndepVarComp('yield_stress', 0.0), promotes=['*'])

        # Pontoons
        self.add('G',                          IndepVarComp('G', 0.0), promotes=['*'])
        
        # LCOE
        self.add('number_of_turbines', IndepVarComp('number_of_turbines', 0, pass_by_obj=True), promotes=['*'])
        self.add('annual_opex',        IndepVarComp('annual_opex', 0.0), promotes=['*']) # TODO: Replace with output connection
        self.add('fixed_charge_rate',  IndepVarComp('fixed_charge_rate', 0.0), promotes=['*'])
        self.add('discount_rate',      IndepVarComp('discount_rate', 0.0), promotes=['*'])

        # Connect all input variables from all models
        self.connect('water_depth', ['waterD', 'sea_depth'])
        self.connect('hub_height', 'hubH')
        self.connect('tower_outer_diameter', 'towerD', src_indices=[0])
        
        self.connect('wind_beta', 'beta')
        self.connect('mean_current_speed', 'Uc')
        
        self.connect('project_lifetime', ['struc.lifetime','projLife'])
        #self.connect('slope_SN', 'm_SN')
        #self.connect('compute_shear', 'shear')
        #self.connect('compute_stiffnes', 'geom')
        #self.connect('frame3dd_matrix_method', 'Mmethod')
        #self.connect('shift_value', 'shift')
        # TODO:
        #self.connect('number_of_modes', ['nM', 'nF'])
        #self.connect('frame3dd_convergence_tolerance', 'tol')
        #self.connect('lumped_mass_matrix', 'lump')
        #self.connect('stress_standard_value', 'DC')
        
        self.connect('rna.loads.top_F', 'rna_force')
        self.connect('rna.loads.top_M', 'rna_moment')
        self.connect('rna.rna_I_TT', 'rna_I')
        self.connect('rna.rna_mass', ['rnaM', 'rna_mass'])
        self.connect('rna.rna_cm', 'rna_cg')
        self.connect('mass_all_blades', 'rna.blades_mass')
        self.connect('I_all_blades', 'rna.blades_I')
        self.connect('hub_mass', 'rna.hub_mass')
        self.connect('nac_mass', 'rna.nac_mass')
        self.connect('hub_cm', 'rna.hub_cm')
        self.connect('nac_cm', 'rna.nac_cm')
        self.connect('hub_I', 'rna.hub_I')
        self.connect('nac_I', 'rna.nac_I')
        self.connect('tilt', 'rna.tilt')
        self.connect('Fxyz_total','rna.loads.F')
        self.connect('Mxyz_total','rna.loads.M')
        self.connect('rna_weightM','rna.rna_weightM')

        
        self.connect('air_density', ['main.windLoads.rho','analysis.rho'])
        self.connect('air_viscosity', ['main.windLoads.mu','analysis.mu'])
        self.connect('water_density', 'water_density')
        self.connect('water_viscosity', 'main.waveLoads.mu')
        self.connect('wave_height', 'Hs')
        self.connect('wave_period', 'T')
        self.connect('wind_reference_speed', 'Uref')
        self.connect('wind_reference_height', ['zref','wind.zref'])
        self.connect('wind_bottom_height', ['z0','wind.z0'])
        self.connect('shearExp', 'wind.shearExp')
        self.connect('morison_mass_coefficient', 'cm')
        
        self.connect('ballast_cost_rate', 'ballCR')
        self.connect('mooring_cost_rate', 'moorCR')
        self.connect('mooring_cost', 'moorCost')
        self.connect('mooring_diameter', 'moorDia')
        self.connect('number_of_mooring_lines', 'moorLines')
        self.connect('material_cost_rate', 'sSteelCR')
        self.connect('main.tapered_column_cost_rate', ['spStifColCR', 'spTapColCR', 'ssStifColCR'])
        self.connect('pontoon_cost_rate', 'ssTrussCR')

        self.connect('nBlades','blade_number')
        self.connect('mass_one_blade', 'blade_mass')
        self.connect('control_maxOmega', 'rotor_omega')
        self.connect('structural_frequencies', 'tower_freq')
        
        self.connect('turbine_cost_kW', 'turbCapEx')
        self.connect('machine_rating', 'turbR')
        self.connect('diameter', 'rotorD')
        self.connect('bladeLength', 'bladeL')
        self.connect('hub_diameter', 'hubD')
        
        # Link outputs from one model to inputs to another
        self.connect('tower_mass', 'towerM')
        self.connect('dummy_mass', 'off.stack_mass_in')

        self.connect('total_cost', 'subTotCost')
        self.connect('total_mass', 'subTotM')
        self.connect('total_bos_cost', 'bos_costs')

        self.connect('number_of_turbines', ['nTurb', 'turbine_number'])
        self.connect('annual_opex', 'avg_annual_opex')
        self.connect('AEP', 'net_aep')
        self.connect('totInstTime', 'construction_time')
        
         # Use complex number finite differences
        typeStr = 'fd'
        formStr = 'central'
        stepVal = 1e-5
        stepStr = 'relative'
        
        self.deriv_options['type'] = typeStr
        self.deriv_options['form'] = formStr
        self.deriv_options['step_size'] = stepVal
        self.deriv_options['step_calc'] = stepStr





"""
test_DriveSE.py

Created by Katherine Dykes on 2014-01-07.
Copyright (c) NREL. All rights reserved.
"""

import unittest
import numpy as np
from math import pi
from commonse.utilities import check_gradient_unit_test

# from drivese.drive_smooth import YawSystemSmooth, BedplateSmooth
from drivese.drive import Drive3pt, Drive4pt, sys_print
from drivese.drivese_components import LowSpeedShaft_drive, Gearbox_drive, MainBearing_drive, SecondBearing_drive, Bedplate_drive, YawSystem_drive, LowSpeedShaft_drive3pt, \
    LowSpeedShaft_drive4pt, Transformer_drive, HighSpeedSide_drive, Generator_drive, NacelleSystemAdder_drive, AboveYawMassAdder_drive, RNASystemAdder_drive
from drivese.hub import Hub_System_Adder_drive, Hub_drive, PitchSystem_drive, Spinner_drive


# Hub Components
# commented out since Hub_system_Adder_drive was implemented instead of Hub_SE
# class Test_HubSE(unittest.TestCase):

#     def setUp(self):

#         self.hub = Hub_System_Adder_drive()

#         self.hub.rotor_diameter = 126.0 # m
#         self.hub.blade_number  = 3
#         self.hub.blade_root_diameter   = 3.542
#         self.hub.machine_rating = 5000.0
#         # self.hub.blade_mass = 17740.0 # kg
    
#         AirDensity= 1.225 # kg/(m^3)
#         Solidity  = 0.0517
#         RatedWindSpeed = 11.05 # m/s
#         self.hub.rotor_bending_moment = (3.06 * pi / 8) * AirDensity * (RatedWindSpeed ** 2) * (Solidity * (self.hub.rotor_diameter ** 3)) / self.hub.blade_number
    
#         self.hub.gamma = 5.0
#         self.hub.MB1_location = np.array([-3.2, 0.0, 1.0])

#     def test_functionality(self):
        
#         self.hub.run()
        
#         self.assertEqual(round(self.hub.hub_system_mass,1), 45025.7)

class Test_Hub(unittest.TestCase):

    def setUp(self):

        self.hub = Hub_drive()

        # self.hub.rotor_diameter = 126.0 # m
        self.hub.blade_number  = 3
        self.hub.blade_root_diameter   = 3.542
        self.hub.machine_rating = 5000.0

        # AirDensity= 1.225 # kg/(m^3)
        # Solidity  = 0.0517
        # RatedWindSpeed = 11.05 # m/s
        # self.hub.rotor_bending_moment = (3.06 * pi / 8) * AirDensity * (RatedWindSpeed ** 2) * (Solidity * (self.hub.rotor_diameter ** 3)) / self.hub.blade_number

    def test_functionality(self):
        
        self.hub.run()
        
        self.assertEqual(round(self.hub.mass,1), 29852.8)


class Test_PitchSystem(unittest.TestCase):

    def setUp(self):

        self.pitch = PitchSystem_drive()

        self.pitch.blade_mass = 17740.0 # kg
        self.pitch.rotor_diameter = 126.0 # m
        self.pitch.blade_number  = 3
        self.pitch.hub_diameter   = 3.542

        AirDensity= 1.225 # kg/(m^3)
        Solidity  = 0.0517
        RatedWindSpeed = 11.05 # m/s
        self.pitch.rotor_bending_moment = (3.06 * pi / 8) * AirDensity * (RatedWindSpeed ** 2) * (Solidity * (self.pitch.rotor_diameter ** 3)) / self.pitch.blade_number

        self.pitch.gamma = 5.0
        self.pitch.MB1_location = np.array([-3.2, 0.0, 1.0])


    def test_functionality(self):
        
        self.pitch.run()
        
        self.assertEqual(round(self.pitch.mass,1), 13362.4)

class Test_Spinner(unittest.TestCase):

    def setUp(self):

        self.spinner = Spinner_drive()

        self.spinner.rotor_diameter = 126.0 # m
        self.spinner.hub_diameter   = 3.542

        self.spinner.gamma = 5.0
        self.spinner.MB1_location = np.array([-3.2, 0.0, 1.0])

    def test_functionality(self):
        
        self.spinner.run()
        
        self.assertEqual(round(self.spinner.mass,1), 1810.5)


# Nacelle Components

class Test_Drive3pt(unittest.TestCase):

    def setUp(self):

        self.nace = Drive3pt()

        self.nace.rotor_diameter = 126.0 # m
        self.nace.rotor_speed = 12.1 # #rpm m/s
        self.nace.machine_rating = 5000.0
        self.nace.DrivetrainEfficiency = 0.95
        self.nace.rotor_torque =  1.5 * (self.nace.machine_rating * 1000 / self.nace.DrivetrainEfficiency) / (self.nace.rotor_speed * (pi / 30)) # 
        self.nace.rotor_thrust = 599610.0 # N
        self.nace.rotor_mass = 0.0 #accounted for in F_z # kg
        self.nace.rotor_speed = 12.1 #rpm
        self.nace.rotor_bending_moment = -16665000.0 # Nm same as rotor_bending_moment_y
        self.nace.rotor_bending_moment_x = 330770.0# Nm
        self.nace.rotor_bending_moment_y = -16665000.0 # Nm
        self.nace.rotor_bending_moment_z = 2896300.0 # Nm
        self.nace.rotor_force_x = 599610.0 # N
        self.nace.rotor_force_y = 186780.0 # N
        self.nace.rotor_force_z = -842710.0 # N
    
        # NREL 5 MW Drivetrain variables
        self.nace.drivetrain_design = 'geared' # geared 3-stage Gearbox with induction generator machine
        self.nace.machine_rating = 5000.0 # kW
        self.nace.gear_ratio = 96.76 # 97:1 as listed in the 5 MW reference document
        self.nace.gear_configuration = 'eep' # epicyclic-epicyclic-parallel
        self.nace.crane = True # onboard crane present
        self.nace.shaft_angle = 5.0 #deg
        self.nace.shaft_ratio = 0.10
        self.nace.Np = [3,3,1]
        self.nace.ratio_type = 'optimal'
        self.nace.shaft_type = 'normal'
        self.nace.uptower_transformer=True
        self.nace.shrink_disc_mass = 333.3*self.nace.machine_rating/1000.0 # estimated
        self.nace.carrier_mass = 8000.0 # estimated
        self.nace.mb1Type = 'SRB'
        self.nace.mb2Type = 'SRB'
        self.nace.flange_length = 0.5
        self.nace.overhang = 5.0
        self.nace.L_rb = 1.912 # length from hub center to main bearing, leave zero if unknow
    
        self.nace.check_fatigue = 0 #0 if no fatigue check, 1 if parameterized fatigue check, 2 if known loads inputs
    
        # NREL 5 MW Tower Variables
        self.nace.tower_top_diameter = 3.78 # m

    def test_functionality(self):
        
        self.nace.run()
        
        self.assertEqual(round(self.nace.nacelle_mass,1), 191196.9)

class Test_Drive4pt(unittest.TestCase):

    def setUp(self):

        self.nace = Drive4pt()
        self.nace.rotor_diameter = 126.0 # m
        self.nace.rotor_speed = 12.1 # #rpm m/s
        self.nace.machine_rating = 5000.0
        self.nace.DrivetrainEfficiency = 0.95
        self.nace.rotor_torque =  1.5 * (self.nace.machine_rating * 1000 / self.nace.DrivetrainEfficiency) / (self.nace.rotor_speed * (pi / 30)) # 6.35e6 #4365248.74 # Nm
        self.nace.rotor_thrust = 599610.0 # N
        self.nace.rotor_mass = 0.0 #accounted for in F_z # kg
        self.nace.rotor_speed = 12.1 #rpm
        self.nace.rotor_bending_moment = -16665000.0 # Nm same as rotor_bending_moment_y
        self.nace.rotor_bending_moment_x = 330770.0# Nm
        self.nace.rotor_bending_moment_y = -16665000.0 # Nm
        self.nace.rotor_bending_moment_z = 2896300.0 # Nm
        self.nace.rotor_force_x = 599610.0 # N
        self.nace.rotor_force_y = 186780.0 # N
        self.nace.rotor_force_z = -842710.0 # N
    
        # NREL 5 MW Drivetrain variables
        self.nace.drivetrain_design = 'geared' # geared 3-stage Gearbox with induction generator machine
        self.nace.machine_rating = 5000.0 # kW
        self.nace.gear_ratio = 96.76 # 97:1 as listed in the 5 MW reference document
        self.nace.gear_configuration = 'eep' # epicyclic-epicyclic-parallel
        self.nace.crane = True # onboard crane present
        self.nace.shaft_angle = 5.0 #deg
        self.nace.shaft_ratio = 0.10
        self.nace.Np = [3,3,1]
        self.nace.ratio_type = 'optimal'
        self.nace.shaft_type = 'normal'
        self.nace.uptower_transformer=False
        self.nace.shrink_disc_mass = 333.3*self.nace.machine_rating/1000.0 # estimated
        self.nace.carrier_mass = 8000.0 # estimated
        self.nace.mb1Type = 'CARB'
        self.nace.mb2Type = 'SRB'
        self.nace.flange_length = 0.5 #m
        self.nace.overhang = 5.0
        self.nace.gearbox_cm = 0.1
        self.nace.hss_length = 1.5
        self.nace.L_rb = 1.912 # length from hub center to main bearing, leave zero if unknown

        self.nace.check_fatigue = 0 #0 if no fatigue check, 1 if parameterized fatigue check, 2 if known loads inputs

        self.nace.tower_top_diameter = 3.78 # m

    def test_functionality(self):
        
        self.nace.run()
        #sys_print(self.nace)
        
        self.assertEqual(round(self.nace.nacelle_mass,1), 190223.5)

'''
class Test_LowSpeedShaft(unittest.TestCase):

    def setUp(self):

        self.lss = LowSpeedShaft()

        self.lss.rotor_diameter = 126. # rotor diameter [m]
        rotor_speed = 12.1
        DrivetrainEfficiency = 0.95
        machine_rating = 5000.0
        self.lss.rotor_torque = 1.5 * (machine_rating * 1000. / DrivetrainEfficiency) / (rotor_speed * (pi / 30.))
        self.lss.rotor_mass = 142585.75 # rotor mass [kg]

    def test_functionality(self):
        
        self.lss.run()
        
        self.assertEqual(round(self.lss.mass,1), 42381.5)

class Test_MainBearing(unittest.TestCase):

    def setUp(self):

        self.mb = MainBearing()

        self.mb.lss_design_torque = 18691533.1165
        self.mb.lss_diameter = 1.2115
        self.mb.lss_mass = 42381.5447
        self.mb.rotor_speed = 12.1
        self.mb.rotor_diameter = 126.0

    def test_functionality(self):
        
        self.mb.run()
        
        self.assertEqual(round(self.mb.mass,1), 7348.5)

class Test_SecondBearing(unittest.TestCase):

    def setUp(self):

        self.sb = MainBearing()

        self.sb.lss_design_torque = 18691533.1165
        self.sb.lss_diameter = 1.2115
        self.sb.lss_mass = 42381.5447
        self.sb.rotor_speed = 12.1
        self.sb.rotor_diameter = 126.0

    def test_functionality(self):
        
        self.sb.run()
        
        self.assertEqual(round(self.sb.mass,1), 7348.5)
'''

class Test_Gearbox_drive(unittest.TestCase):

    def setUp(self):

        self.gbx = Gearbox_drive()

        self.gbx.gear_ratio = 96.76 # 97:1 as listed in the 5 MW reference document
        self.gbx.gear_configuration = 'eep' # epicyclic-epicyclic-parallel
        self.gbx.Np = [3,3,1]
        self.gbx.ratio_type = 'optimal'
        self.gbx.shaft_type = 'normal'
        self.gbx.cm_input = 0.1
        self.gbx.hss_length = 1.5

        self.gbx.rotor_diameter = 126.0 # m
        self.gbx.rotor_speed = 12.1 # #rpm m/s
        self.gbx.machine_rating = 5000.0
        self.gbx.DrivetrainEfficiency = 0.95
        self.gbx.rotor_torque =  1.5 * (self.gbx.machine_rating * 1000 / self.gbx.DrivetrainEfficiency) / (self.gbx.rotor_speed * (pi / 30)) # 6.35e6 #4365248.74 # Nm

    def test_functionality(self):
        
        self.gbx.run()
        
        self.assertEqual(round(self.gbx.mass,1), 55658.3)

class Test_HighSpeedSide(unittest.TestCase):

    def setUp(self):

        self.hss = HighSpeedSide_drive()

        self.hss.gear_ratio = 96.76
        self.hss.rotor_diameter = 126. # rotor diameter [m]
        rotor_speed = 12.1
        DrivetrainEfficiency = 0.95
        machine_rating = 5000.0
        self.hss.rotor_torque = 1.5 * (machine_rating * 1000. / DrivetrainEfficiency) / (rotor_speed * (pi / 30.))
        self.hss.lss_diameter = 1.2115

        self.hss.gearbox_length = 1.512
        self.hss.gearbox_height = 1.89
        self.hss.gearbox_cm = np.array([0.1, 0., 0.756])
        self.hss.length_in = 1.5

    def test_functionality(self):
        
        self.hss.run()
        
        self.assertEqual(round(self.hss.mass,1), 2414.7)

class Test_Generator(unittest.TestCase):

    def setUp(self):

        self.gen = Generator_drive()

        self.gen.gear_ratio = 96.76
        self.gen.rotor_diameter = 126. # rotor diameter [m]
        self.gen.machine_rating = 5000.
        self.gen.drivetrain_design = 'geared'

        self.gen.highSpeedSide_length = 1.5
        self.gen.highSpeedSide_cm = np.array([1.606, 0., 1.134])
        self.gen.rotor_speed = 12.1

    def test_functionality(self):
        
        self.gen.run()
        
        self.assertEqual(round(self.gen.mass,1), 16699.9)


class Test_Bedplate(unittest.TestCase):

    def setUp(self):

        self.bpl = Bedplate_drive()

        self.bpl.gbx_length = 1.512
        self.bpl.gbx_location = 0.0
        self.bpl.gbx_mass = 0.0
        self.bpl.hss_location = 1.606
        self.bpl.hss_mass = 2414.6771802
        self.bpl.generator_location  = 4.057
        self.bpl.generator_mass  = 16699.851325
        self.bpl.lss_location  = -2.52145326316
        self.bpl.lss_mass  = 18564.9450561
        self.bpl.lss_length  = 3.165
        self.bpl.mb1_location = -3.18633453315
        self.bpl.FW_mb1 = 0.25
        self.bpl.mb1_mass  = 4438.51851852
        self.bpl.mb2_location = -0.745657522828
        self.bpl.mb2_mass = 1406.85185185
        self.bpl.transformer_mass = 0.0
        self.bpl.transformer_location = 0.0
        self.bpl.tower_top_diameter = 3.78
        self.bpl.rotor_diameter = 126.0
        self.bpl.machine_rating = 5000.0
        self.bpl.rotor_mass  = 0.0
        self.bpl.rotor_bending_moment_y = -16665000.0
        self.bpl.rotor_force_z  = -842710.0
        self.bpl.flange_length  = 0.5
        self.bpl.L_rb = 1.912
        self.bpl.overhang  = 5.0
    
        #parameters
        self.bpl.uptower_transformer = True

    def test_functionality(self):
        
        self.bpl.run()
        
        self.assertEqual(round(self.bpl.mass,1), 64485.7)

class Test_YawSystem(unittest.TestCase):

    def setUp(self):

        self.yaw = YawSystem_drive()

        #variables
        self.yaw.rotor_diameter = 126.
        self.yaw.rotor_thrust = 599610.0
        self.yaw.tower_top_diameter = 3.78
        self.yaw.above_yaw_mass = 164945.719525
        self.yaw.bedplate_height = 1.4496
    
        #parameters
        self.yaw.yaw_motors_number = 0

    def test_functionality(self):
        
        self.yaw.run()
        
        self.assertEqual(round(self.yaw.mass,1), 6044.7)

class Test_Transformer(unittest.TestCase):

    def setUp(self):

        self.trans = Transformer_drive()

        self.trans.machine_rating = 5000.0
        self.trans.uptower_transformer = False
        self.trans.tower_top_diameter =3.78
        self.trans.rotor_mass = 0.0
        self.trans.overhang = 5.0
        self.trans.generator_cm = np.array([4.057, 0.0, 1.134])
        self.trans.rotor_diameter = 126.0
        self.trans.RNA_mass = 217013.124235
        self.trans.RNA_cm = -2.64480400163

    def test_functionality(self):
        
        self.trans.run()
        
        self.assertEqual(round(self.trans.mass,1), 0.0)


'''
class Test_AboveYawMassAdder(unittest.TestCase):

    def setUp(self):

        self.yawadder = AboveYawMassAdder()

        self.yawadder.machine_rating = 5000.0
        self.yawadder.lss_mass = 42381.5
        self.yawadder.main_bearing_mass = 14696.2/2.
        self.yawadder.second_bearing_mass = 14696.2/2.
        self.yawadder.gearbox_mass = 48664.7
        self.yawadder.hss_mass = 2414.7
        self.yawadder.generator_mass = 16699.9
        self.yawadder.bedplate_mass = 108512.5
        self.yawadder.bedplate_length = 10.4006
        self.yawadder.bedplate_width = 5.20032
        self.yawadder.crane = True

    def test_functionality(self):
        
        self.yawadder.run()
        
        self.assertEqual(round(self.yawadder.above_yaw_mass,1), 259430.9)


class Test_NacelleSystemAdder(unittest.TestCase):

    def setUp(self):

        self.nac = NacelleSystemAdder()

    def test_functionality(self):

        self.nac.above_yaw_mass = 259430.9
        self.nac.yawMass = 13789.0
        self.nac.machine_rating = 5000.0
        self.nac.lss_mass = 42381.5
        self.nac.main_bearing_mass = 14696.2/2.
        self.nac.second_bearing_mass = 14696.2/2.
        self.nac.gearbox_mass = 48664.7
        self.nac.hss_mass = 2414.7
        self.nac.generator_mass = 16699.9
        self.nac.bedplate_mass = 108512.5
        self.nac.mainframe_mass = 125076.5
        self.nac.lss_cm = np.random.rand(3)  # np.array([-2.0, 1.0, 1.0])
        self.nac.main_bearing_cm = np.random.rand(3)  # np.array([-2.0, 1.0, 1.0])
        self.nac.second_bearing_cm = np.random.rand(3)  # np.array([-2.0, 1.0, 1.0])
        self.nac.gearbox_cm = np.random.rand(3)  # np.array([-2.0, 1.0, 1.0])
        self.nac.hss_cm = np.random.rand(3)  # np.array([-2.0, 1.0, 1.0])
        self.nac.generator_cm = np.random.rand(3)  # np.array([-2.0, 1.0, 1.0])
        self.nac.bedplate_cm = np.random.rand(3)  # np.array([-2.0, 1.0, 1.0])
        self.nac.lss_I = np.random.rand(3)  # np.array([1000., 1000., 1000.])
        self.nac.main_bearing_I = np.random.rand(3)  # np.array([1000., 1000., 1000.])
        self.nac.second_bearing_I = np.random.rand(3)  # np.array([1000., 1000., 1000.])
        self.nac.gearbox_I = np.random.rand(3)  # np.array([1000., 1000., 1000.])
        self.nac.hss_I = np.random.rand(3)  # np.array([1000., 1000., 1000.])
        self.nac.generator_I = np.random.rand(3)  # np.array([1000., 1000., 1000.])
        self.nac.bedplate_I = np.random.rand(3)  # np.array([1000., 1000., 1000.])

        self.nac.run()

        self.assertEqual(round(self.nac.nacelle_mass,1), 273219.9)

'''

# Gradient tests for drive smooth
# TODO: no unit tests since drive_smooth needs updating
# class TestBearingSmooth(unittest.TestCase):

#     def test_gradient(self):
#         comp = BearingSmooth()
#         comp.bearing_type = 'SRB'
#         comp.lss_diameter = 0.721049014299
#         comp.rotor_diameter = 125.740528176
#         comp.bearing_switch = 'main'

#         check_gradient_unit_test(self, comp)


# class TestYawSystemSmooth(unittest.TestCase):

#     def test_gradient(self):
#         comp = YawSystemSmooth()
#         comp.rotor_diameter = 125.740528176
#         comp.tower_top_diameter = 3.87

#         check_gradient_unit_test(self, comp)


# class TestBedplateSmooth(unittest.TestCase):

#     def test_gradient(self):
#         comp = BedplateSmooth()
#         comp.hss_location = 0.785878301101
#         comp.hss_mass = 2288.26758514
#         comp.generator_location = 1.5717566022
#         comp.generator_mass = 16699.851325
#         comp.lss_location = -3.14351320441
#         comp.lss_mass = 12546.3193435
#         comp.mb1_location = -1.25740528176
#         comp.mb1_mass = 3522.06734168
#         comp.mb2_location = -4.40091848617
#         comp.mb2_mass = 5881.81400444
#         comp.tower_top_diameter = 3.87
#         comp.rotor_diameter = 125.740528176
#         comp.machine_rating = 5000.0
#         comp.rotor_mass = 93910.5225629
#         comp.rotor_bending_moment_y = -2325000.0
#         comp.rotor_force_z = -921262.226342
#         comp.h0_rear = 1.35
#         comp.h0_front = 1.7

#         check_gradient_unit_test(self, comp)
        
if __name__ == "__main__":
    unittest.main()
    


'''
Components for the new spherical hub model - July 2019
  Sph_Hub, PitchSystem, Sph_Spinner, Hub_System_Adder, Hub_Mass_Adder, Hub_CM_Adder are now imported 
  from sph_hubse_components.py (instead of hubse_components.py)

'''

import numpy as np

from openmdao.api import ExplicitComponent, Group, Problem, IndepVarComp
from wisdem.drivetrainse.hubse_components import Hub, PitchSystem, Spinner, Hub_System_Adder, Hub_Mass_Adder, Hub_CM_Adder

#-------------------------------------------------------------------------

class Hub_System_Adder_OM(ExplicitComponent):
    ''' 
    Compute hub mass, cm, and I

    NOTE: THIS COMPONENT DOES NOT PLAY WELL WITH DRIVESE BECAUSE IT CREATES CIRCULAR REFERENCES!
    (requires main bearing location, but outputs from here are required to determine MB1_Location in drivese_omdao.py)
    '''
    def initialize(self):
        self.options.declare('debug', default=False)
        
    def setup(self):
        # variables
        self.add_input('rotor_diameter',    val=0.0,                     units='m',   desc='rotor diameter')
        self.add_input('distance_hub2mb',   val=0.0,                     units='m',   desc='distance between hub center and upwind main bearing')
        self.add_input('shaft_angle',       val=0.0,                     units='rad', desc='shaft angle')
        self.add_input('MB1_location',      val=np.zeros(3), shape=(3,), units='m',   desc='center of mass of main bearing in [x,y,z] for an arbitrary coordinate system')
        self.add_input('hub_mass',          val=0.0,                     units='kg',  desc='mass of Hub')
        self.add_input('hub_diameter',      val=0.03,                    units='m',   desc='hub diameter')
        self.add_input('hub_thickness',     val=0.0,                     units='m',   desc='hub thickness')
        self.add_input('pitch_system_mass', val=0.0,                     units='kg',  desc='mass of Pitch System')
        self.add_input('spinner_mass',      val=0.0,                     units='kg',  desc='mass of spinner')
        self.add_input('blade_mass',        val=0.0,                     units='kg',  desc='mass of one blade')
        self.add_discrete_input('number_of_blades', val=0, desc='number of blades on the rotor')
        
        # outputs
        self.add_output('hub_system_cm',   val=np.zeros(3),             units='m',  desc='center of mass of the hub relative to tower to in yaw-aligned c.s.')
        self.add_output('hub_system_I',    val=np.zeros(6),             units='kg*m**2', desc='mass moments of Inertia of hub [Ixx, Iyy, Izz, Ixy, Ixz, Iyz] around its center of mass in yaw-aligned c.s.')
        self.add_output('hub_I',           val=np.zeros(3),             units='kg*m**2', desc='Hub inertia about rotor axis (does not include pitch and spinner masses)')
        self.add_output('hub_system_mass', val=0.0,                     units='kg', desc='mass of hub system')
        self.add_output('rotor_mass',      val=0.0,                     units='kg', desc='mass of rotor')

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
    
        hub_adder = Hub_System_Adder(discrete_inputs['number_of_blades'], debug=self.options['debug'])

        (outputs['rotor_mass'], outputs['hub_system_mass'], outputs['hub_system_cm'], outputs['hub_system_I'], outputs['hub_I']) \
            = hub_adder.compute(inputs['rotor_diameter'], inputs['blade_mass'], inputs['distance_hub2mb'], inputs['shaft_angle'], inputs['MB1_location'], \
                                inputs['hub_mass'], inputs['hub_diameter'], inputs['hub_thickness'], inputs['pitch_system_mass'], inputs['spinner_mass'])        


#-------------------------------------------------------------------------

class Hub_Mass_Adder_OM(ExplicitComponent):
    ''' 
    Compute hub mass and I
    Excluding cm here, because it has a dependency on main bearing location, which can only be calculated once the full rotor mass is set
    '''
    def initialize(self):
        self.options.declare('debug', default=False)
        
    def setup(self):
        # variables
        self.add_input('hub_mass',          val=0.0,  units='kg', desc='mass of Hub')
        self.add_input('hub_diameter',      val=0.03, units='m',  desc='hub diameter')
        self.add_input('hub_thickness',     val=0.0,  units='m',  desc='hub thickness')
        self.add_input('pitch_system_mass', val=0.0,  units='kg', desc='mass of Pitch System')
        self.add_input('spinner_mass',      val=0.0,  units='kg', desc='mass of spinner')
        self.add_input('blade_mass',        val=0.0,  units='kg', desc='mass of one blade')
        self.add_discrete_input('number_of_blades', val=0, desc='number of blades on the rotor')

        # outputs
        self.add_output('hub_system_I',    val=np.zeros(6),             units='kg*m**2', desc='mass moments of Inertia of hub [Ixx, Iyy, Izz, Ixy, Ixz, Iyz] around its center of mass in yaw-aligned c.s.')
        self.add_output('hub_system_mass', val=0.0,                     units='kg', desc='mass of hub system')
        self.add_output('rotor_mass',      val=0.0,                     units='kg', desc='mass of rotor')
        self.add_output('hub_I',           val=np.zeros(3),             units='kg*m**2', desc='Hub inertia about rotor axis (does not include pitch and spinner masses)')


    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):

        hub_adder = Hub_Mass_Adder(discrete_inputs['number_of_blades'], debug=self.options['debug'])
        
        (outputs['rotor_mass'], outputs['hub_system_mass'], outputs['hub_system_I'], outputs['hub_I'])\
            = hub_adder.compute(inputs['blade_mass'], inputs['hub_mass'], inputs['hub_diameter'], 
                                inputs['hub_thickness'], inputs['pitch_system_mass'], inputs['spinner_mass'])        
        

#-------------------------------------------------------------------------

class Hub_CM_Adder_OM(ExplicitComponent):
    ''' 
    Compute hub cm
    Separating cm here, because it has a dependency on main bearing location, which can only be calculated once the full rotor mass is set
    '''

    def setup(self):
        # variables
        self.add_input('rotor_diameter',  val=0.0,         units='m',   desc='rotor diameter')
        self.add_input('distance_hub2mb', val=0.0,         units='m',   desc='distance between hub center and upwind main bearing')
        self.add_input('shaft_angle',     val=0.0,         units='rad', desc='shaft angle')
        self.add_input('MB1_location',    val=np.zeros(3), units='m',   desc='center of mass of main bearing in [x,y,z] for an arbitrary coordinate system')

        # outputs
        self.add_output('hub_system_cm', val=np.zeros(3), units='m',  desc='center of mass of the hub relative to tower to in yaw-aligned c.s.')

    def compute(self, inputs, outputs):
    
        hub_adder = Hub_CM_Adder()

        outputs['hub_system_cm'] = hub_adder.compute(inputs['rotor_diameter'], inputs['distance_hub2mb'],
                                                     inputs['shaft_angle'], inputs['MB1_location'])
        
    
# -------------------------------------------------

class Hub_OM(ExplicitComponent):
    ''' Hub class    
          The Hub class is used to represent the hub component of a wind turbine. 
          It contains the general properties for a wind turbine component as well as additional design load and dimentional attributes as listed below.
          It contains an update method to determine the mass, mass properties, and dimensions of the component.            
    '''
    def initialize(self):
        self.options.declare('debug', default=False)
        
    def setup(self):
        # variables
        self.add_input('blade_root_diameter', val=0.0, units='m',   desc='blade root diameter')
        self.add_input('machine_rating',      val=0.0, units='kW',  desc='machine rating of turbine')
        self.add_input('rotor_rpm',           val=0.0, units='rpm', desc='RPM at rated power')
        self.add_input('blade_mass',          val=0.0, units='kg',  desc='mass of one blade')
        self.add_input('rotor_diameter',      val=0.0, units='m',   desc='diameter of rotor')
        self.add_input('blade_length',        val=0.0, units='m',   desc='length of blade')
        self.add_discrete_input('number_of_blades', val=0, desc='number of blades on the rotor')

        # outputs
        self.add_output('hub_diameter',  val=0.0, units='m',  desc='hub diameter')
        self.add_output('hub_thickness', val=0.0, units='m',  desc='hub thickness')
        self.add_output('hub_mass',      val=0.0, units='kg', desc='overall component mass')
        self.add_output('hub_cm',        val=0.0, units='m',  desc='hub center of mass')
        self.add_output('hub_cost',      val=0.0, units='USD',  desc='hub cost')

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
    
        myhub = Hub(discrete_inputs['number_of_blades'], debug=self.options['debug'])

        (outputs['hub_mass'], outputs['hub_diameter'], outputs['hub_cm'], outputs['hub_cost'], outputs['hub_thickness']) \
            = myhub.compute(inputs['blade_root_diameter'], inputs['rotor_rpm'],
                            inputs['blade_mass'], inputs['rotor_diameter'], inputs['blade_length']) 


class PitchSystem_OM(ExplicitComponent):
    '''
     PitchSystem class
      The PitchSystem class is used to represent the pitch system of a wind turbine.
      It contains the general properties for a wind turbine component as well as additional design load and dimentional attributes as listed below.
      It contains an update method to determine the mass, mass properties, and dimensions of the component.
    '''
    def initialize(self):
        self.options.declare('debug', default=False)
        
    def setup(self):
        # variables
        self.add_input('blade_mass',             val=0.0, units='kg',  desc='mass of one blade')
        self.add_input('rotor_bending_moment_y', val=0.0, units='N*m', desc='flapwise bending moment at blade root')
        self.add_discrete_input('number_of_blades', val=0, desc='number of blades on the rotor')

        # outputs
        self.add_output('pitch_system_mass',     val=0.0, units='kg',  desc='overall component mass')

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
    
        mypitch = PitchSystem(discrete_inputs['number_of_blades'], debug=self.options['debug'])
        
        outputs['pitch_system_mass'] = mypitch.compute(inputs['blade_mass'], inputs['rotor_bending_moment_y'])


#-------------------------------------------------------------------------

class Spinner_OM(ExplicitComponent):
    '''
       Spinner class
          The SpinnerClass is used to represent the spinner of a wind turbine.
          It contains the general properties for a wind turbine component as well as additional design load and dimentional attributes as listed below.
          It contains an update method to determine the mass, mass properties, and dimensions of the component.
    '''

    def initialize(self):
        self.options.declare('debug', default=False)
        
    def setup(self):
        # variables
        #self.add_input('rotor_diameter', val=0.0, units='m', desc='rotor diameter')
        self.add_input('blade_root_diameter', val=0.0, units='m',  desc='blade root diameter')
        self.add_discrete_input('number_of_blades', val=0, desc='number of blades on the rotor')

        # outputs
        self.add_output('spinner_mass',       val=0.0, units='kg', desc='overall component mass')

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        myspinner = Spinner(discrete_inputs['number_of_blades'], debug=self.options['debug'])

        (outputs['spinner_mass'], _, _)  = myspinner.compute(inputs['blade_root_diameter']) 
        

#%%-----------------------

class HubSE(Group):
    def initialize(self):
        self.options.declare('debug', default=False)
        self.options.declare('mass_only', default=False)
        self.options.declare('topLevelFlag', default=True)
        
    def setup(self):
        debug = self.options['debug']

        '''
           HubSE class
              The HubSE class is used to represent the hub system of a wind turbine. 
              HubSE integrates the hub, pitch system and spinner / nose cone components for the hub system.
        '''
        self.add_subsystem('hub',         Hub_OM(        debug=debug), promotes=['*'])
        self.add_subsystem('pitchSystem', PitchSystem_OM(debug=debug), promotes=['*'])
        self.add_subsystem('spinner',     Spinner_OM(    debug=debug), promotes=['*'])
        if self.options['mass_only']:
            self.add_subsystem('adder',   Hub_Mass_Adder_OM(debug=debug), promotes=['*'])
        else:
            self.add_subsystem('adder',   Hub_System_Adder_OM(debug=debug), promotes=['*'])
        
        if self.options['topLevelFlag']:
            sharedIndeps = IndepVarComp()
            sharedIndeps.add_output('rotor_diameter', 0.0, units='m')
            sharedIndeps.add_discrete_output('number_of_blades', 0)
            sharedIndeps.add_output('blade_root_diameter', 0.0, units='m')
            sharedIndeps.add_output('blade_mass', 0.0, units='kg')
            sharedIndeps.add_output('machine_rating', 0.0, units='kW')
            sharedIndeps.add_output('shaft_angle', 0.0, units='rad')
            sharedIndeps.add_output('rotor_rpm', 0.0, units='rpm')
            self.add_subsystem('sharedIndeps', sharedIndeps, promotes=['*'])

        
#%%----------------------------
        
        #-------------------------------------------------------------------------
        # Examples based on reference turbines including the NREL 5 MW, WindPACT 1.5 MW and the GRC 750 kW system.

def example_5MW_4pt(debug=False):
    
    # simple test of module
    
    prob=Problem()
    prob.model = HubSE(debug=debug)
    '''
    recorder = HDF5Recorder('dump5MW4pt.h5')
    recorder.options['record_params'] = True
    recorder.options['record_metadata'] = True
    prob.driver.add_recorder(recorder)
    '''
    
    prob.setup()
    blade_number = 3
    prob['number_of_blades'] = blade_number
    prob['rotor_diameter'] = 126.0  # m
    prob['blade_root_diameter'] = 3.542
    prob['machine_rating'] = 5000.0
    prob['blade_mass'] = 17740.0  # kg
    prob['shaft_angle'] = np.radians(5)
    prob['rotor_rpm'] = 12.1
    #prob['distance_hub2mb'] = 0.0
    #prob['MB1_location'] = [0.0, 0.0, 0.0]

    AirDensity = 1.225  # kg/(m^3)
    Solidity = 0.0517
    RatedWindSpeed = 11.05  # m/s
    prob['rotor_bending_moment_y'] = (3.06 * np.pi / 8) * AirDensity \
         * (RatedWindSpeed ** 2) * (Solidity * (prob['rotor_diameter'] ** 3)) / blade_number

    prob.run_driver()

    print("NREL 5 MW spherical-hub turbine test")
    prob.model.list_inputs(units=True)#values = False, hierarchical=False)
    prob.model.list_outputs(units=True)#values = False, hierarchical=False)    

#%%----------------------------
        
def example_BAR_4pt():
    
    # created 2019 05 03 by GNS - copied and modified example_5MW_4pt()
    
    blade_number = 3

    prob=Problem()
    prob.model = HubSE(debug=debug)
    prob.setup()

    blade_number = 3
    prob['number_of_blades'] = blade_number
    prob['rotor_diameter'] = 200.0  # m
    prob['blade_root_diameter'] = 4.5
    prob['machine_rating'] = 5000.0
    prob['blade_mass'] = 61400.0  # kg
    prob['shaft_angle'] = np.radians(5)

    AirDensity = 1.225  # kg/(m^3)
    Solidity = 0.0517
    RatedWindSpeed = 11.05  # m/s
    prob['rotor_bending_moment_y'] = (3.06 * np.pi / 8) * AirDensity \
         * (RatedWindSpeed ** 2) * (Solidity * (prob['rotor_diameter'] ** 3)) / blade_number

    prob.run_driver()

    print("NREL BAR turbine test")
    prob.model.list_inputs(units=True)#values = False, hierarchical=False)
    prob.model.list_outputs(units=True)#values = False, hierarchical=False)    

'''
# TODO: update other examples for the hub system
def example_1p5MW_4pt():

    # WindPACT 1.5 MW turbine
    prob = Problem()
    prob.model = HubSE()
    prob.setup()

    prob['blade_mass'] = 4470.0
    prob['rotor_diameter'] = 70.0
    prob['blade_number'] = 3
    prob['hub_diameter'] = 3.0
    prob['machine_rating'] = 1500.0
    prob['blade_root_diameter'] = 2.0  # TODO - find actual number

    AirDensity = 1.225
    Solidity = 0.065
    RatedWindSpeed = 12.12
    prob['rotor_bending_moment'] = (3.06 * pi / 8) * AirDensity * (
        RatedWindSpeed ** 2) * (Solidity * (prob['rotor_diameter'] ** 3)) / prob['blade_number']

    prob.run_driver()

    print("WindPACT 1.5 MW turbine test")
    print("Hub Objects")
    print('  hub         {0:8.1f} kg'.format(prob['hub_mass']))  # 31644.47
    print('  pitch mech  {0:8.1f} kg'.format(prob['pitch_system_mass']))  # 17003.98
    print('  nose cone   {0:8.1f} kg'.format(prob['spinner_mass']))  # 1810.50
    # 50458.95
    print('Hub system total {0:8.1f} kg'.format(prob['hub_system_mass']))
    print('    cm {0:6.2f} {1:6.2f} {2:6.2f}'.format(prob['hub_system_cm'][0], prob['hub_system_cm'][1], prob['hub_system_cm'][2]))
    print('    I {0:6.1f} {1:6.1f} {2:6.1f}'.format(prob['hub_system_I'][0], prob['hub_system_I'][1], prob['hub_system_I'][2]))
    print()

def example_750kW_4pt():

    # GRC 750 kW turbine
    prob = Problem()
    prob.model = HubSE()
    prob.setup()

    prob['blade_mass'] = 3400.0
    prob['rotor_diameter'] = 48.2
    prob['blade_number'] = 3
    prob['hub_diameter'] = 3.0
    prob['machine_rating'] = 750.0
    prob['blade_root_diameter'] = 1.0  # TODO - find actual number

    AirDensity = 1.225
    Solidity = 0.07  # uknown value
    RatedWindSpeed = 16.0
    prob['rotor_bending_moment'] = (3.06 * pi / 8) * AirDensity * (
        RatedWindSpeed ** 2) * (Solidity * (prob['rotor_diameter'] ** 3)) / prob['blade_number']

    prob.run_driver()

    print("windpact 750 kW turbine test")
    print("Hub Objects")
    print('  hub         {0:8.1f} kg'.format(prob['hub_mass']))  # 31644.47
    print('  pitch mech  {0:8.1f} kg'.format(prob['pitch_system_mass']))  # 17003.98
    print('  nose cone   {0:8.1f} kg'.format(prob['spinner_mass']))  # 1810.50
    # 50458.95
    print('Hub system total {0:8.1f} kg'.format(prob['hub_system_mass']))
    print('    cm {0:6.2f} {1:6.2f} {2:6.2f}'.format(prob['hub_system_cm'][0], prob['hub_system_cm'][1], prob['hub_system_cm'][2]))
    print('    I {0:6.1f} {1:6.1f} {2:6.1f}'.format(prob['hub_system_I'][0], prob['hub_system_I'][1], prob['hub_system_I'][2]))
    print()

'''

# Main code to run hub system examples
if __name__ == "__main__":

    debug = True
    example_5MW_4pt(debug=debug)
